from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class JSONArrayUnpack(Value):
    """32-bit float JSON array unpacker."""

    arg = binary

    output_dtype = dt.json
    output_shape = rlz.shape_like('arg')


def json_array_unpack(arg: ir.BinaryValue) -> ir.JSONValue:
    """
    Convert an encoded blob representing a 32-bit float vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return JSONArrayUnpack(arg).to_expr()


ir.BinaryValue.json_array_unpack = json_array_unpack


class JSONArrayUnpackI8(JSONArrayUnpack):
    """8-bit integer JSON array unpacker."""


def json_array_unpack_i8(arg: ir.BinaryValue) -> ir.JSONValue:
    """
    Convert an encoded blob representing an 8-bit int vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return JSONArrayUnpackI8(arg).to_expr()


ir.BinaryValue.json_array_unpack_i8 = json_array_unpack_i8


class JSONArrayUnpackI16(JSONArrayUnpack):
    """16-bit integer JSON array unpacker."""


def json_array_unpack_i16(arg: ir.BinaryValue) -> ir.JSONValue:
    """
    Convert an encoded blob representing a 16-bit int vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return JSONArrayUnpackI16(arg).to_expr()


ir.BinaryValue.json_array_unpack_i16 = json_array_unpack_i16


class JSONArrayUnpackI32(JSONArrayUnpack):
    """32-bit integer JSON array unpacker."""


def json_array_unpack_i32(arg: ir.BinaryValue) -> ir.JSONValue:
    """
    Convert an encoded blob representing a 32-bit int vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return JSONArrayUnpackI32(arg).to_expr()


ir.BinaryValue.json_array_unpack_i32 = json_array_unpack_i32


class JSONArrayUnpackI64(JSONArrayUnpack):
    """64-bit integer JSON array unpacker."""


def json_array_unpack_i64(arg: ir.BinaryValue) -> ir.JSONValue:
    """
    Convert an encoded blob representing a 64-bit int vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return JSONArrayUnpackI64(arg).to_expr()


ir.BinaryValue.json_array_unpack_i64 = json_array_unpack_i64


class JSONArrayUnpackF32(JSONArrayUnpack):
    """32-bit float JSON array unpacker."""


def json_array_unpack_f32(arg: ir.BinaryValue) -> ir.JSONValue:
    """
    Convert an encoded blob representing a 32-bit float vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return JSONArrayUnpackF32(arg).to_expr()


ir.BinaryValue.json_array_unpack_f32 = json_array_unpack_f32


class JSONArrayUnpackF64(JSONArrayUnpack):
    """64-bit float JSON array unpacker."""


def json_array_unpack_f64(arg: ir.BinaryValue) -> ir.JSONValue:
    """
    Convert an encoded blob representing a 64-bit float vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return JSONArrayUnpackF64(arg).to_expr()


ir.BinaryValue.json_array_unpack_f64 = json_array_unpack_f64



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONLength(Value):
    """JSON object / array length."""

    arg = rlz.one_of([rlz.json, rlz.string])

    output_dtype = dt.int
    output_shape = rlz.shape_like('arg')


def json_length(arg: ir.JSONValue | ir.StringValue) -> ir.IntegerValue:
    """
    Get the length of the JSON object / array.

    Parameters
    ----------
    arg : JSON or string
        JSON array

    Returns
    -------
    Integer column

    """
    return JSONLength(arg).to_expr()


ir.JSONValue.json_length = json_length
ir.StringValue.json_length = json_length



from __future__ import annotations

from .dot_product import *
from .euclidean_distance import *
from .json_array_pack import *
from .json_array_unpack import *
from .scalar_vector_mul import *
from .vector_add import *
from .vector_elements_sum import *
from .vector_kth_element import *
from .vector_mul import *
from .vector_num_elements import *
from .vector_sort import *
from .vector_sub import *
from .vector_subvector import *
from .vector_sum import *



from __future__ import annotations

import os
from datetime import date
from operator import methodcaller
from typing import Any
from typing import Generator

import ibis
import ibis.expr.datatypes as dt
import pandas as pd
import pandas.testing as tm
import pytest
import sqlalchemy as sa
import sqlalchemy_singlestoredb
from ibis.util import gen_name
from pytest import param


SINGLESTOREDB_DB = os.environ.get('IBIS_TEST_SINGLESTOREDB_DATABASE', 'ibis_testing')

SINGLESTOREDB_TYPES = [
    ('bool', dt.int8),
    ('boolean', dt.int8),
    ('bit', dt.int64),
    ('bit(1)', dt.int64),
    ('bit(9)', dt.int64),
    ('bit(17)', dt.int64),
    ('bit(33)', dt.int64),
    ('tinyint', dt.int8),
    ('tinyint unsigned', dt.uint8),
    ('smallint', dt.int16),
    ('smallint unsigned', dt.uint16),
    ('mediumint', dt.int32),
    ('mediumint unsigned', dt.uint32),
    ('int', dt.int32),
    ('int unsigned', dt.uint32),
    ('integer', dt.int32),
    ('integer unsigned', dt.uint32),
    ('bigint', dt.int64),
    ('bigint unsigned', dt.uint64),
    ('int1', dt.int8),
    ('int1 unsigned', dt.uint8),
    ('int2', dt.int16),
    ('int2 unsigned', dt.uint16),
    ('int3', dt.int32),
    ('int3 unsigned', dt.uint32),
    ('int4 unsigned', dt.uint32),
    ('float', dt.float32),
    ('double', dt.float64),
    ('decimal(10, 0)', dt.Decimal(10, 0)),
    ('decimal(5, 2)', dt.Decimal(5, 2)),
    ('dec', dt.Decimal(10, 0)),
    ('dec(5, 2)', dt.Decimal(5, 2)),
    ('numeric', dt.Decimal(10, 0)),
    ('fixed', dt.Decimal(10, 0)),
    ('date', dt.date),
    ('time', dt.time),
    ('time(6)', dt.time),
    ('datetime', dt.timestamp),
    ('datetime(6)', dt.timestamp),
    ('timestamp', dt.Timestamp('UTC')),
    ('timestamp(6)', dt.Timestamp('UTC')),
    ('year', dt.uint8),
    ('char(32)', dt.string),
    ('binary(42)', dt.binary),
    ('varchar(42)', dt.string),
    ('char byte', dt.binary),
    ('text', dt.string),
    ('varbinary(42)', dt.binary),
    ('longtext', dt.string),
    ('mediumtext', dt.string),
    ('text', dt.string),
    ('tinytext', dt.string),
    ('text(4)', dt.string),
    ('longblob', dt.binary),
    ('mediumblob', dt.binary),
    ('blob', dt.binary),
    ('tinyblob', dt.binary),
    ('json', dt.json),
    ('geographypoint', dt.string),
    ('geography', dt.string),
    ("enum('small', 'medium', 'large')", dt.string),
    # TODO: The SET flag isn't sent to the client
    #   ("set('a', 'b', 'c', 'd')", dt.Array(dt.string)),
    ('bit(7)', dt.int64),
    # ('uuid', dt.string),
]


@pytest.mark.parametrize(
    ('singlestoredb_type', 'expected_type'),
    [
        param(singlestoredb_type, ibis_type, id=singlestoredb_type)
        for singlestoredb_type, ibis_type in SINGLESTOREDB_TYPES
    ],
)
def test_get_schema_from_query(
    con: Any,
    singlestoredb_type: Any,
    expected_type: Any,
) -> None:
    if 'unsigned' in singlestoredb_type and (
        'http://' in str(con.con.url) or 'https://' in str(con.con.url)
    ):
        pytest.skip('HTTP API does not surface unsigned int information')
        return

    tmp = f'tmp_{ibis.util.guid()}'

    try:
        with con.begin() as c:
            c.exec_driver_sql(
                f'CREATE ROWSTORE TABLE {tmp} (x {singlestoredb_type})',
            )
        expected_schema = ibis.schema(dict(x=expected_type))
        t = con.table(tmp)
        result_schema = con._get_schema_using_query(f'SELECT * FROM {tmp}')
        assert t.schema() == expected_schema
        assert result_schema == expected_schema

    finally:
        con.drop_table(tmp, force=True)


@pytest.mark.parametrize('coltype', ['TINYBLOB', 'MEDIUMBLOB', 'BLOB', 'LONGBLOB'])
def test_blob_type(con: Any, coltype: str) -> None:
    tmp = f'tmp_{ibis.util.guid()}'

    try:
        with con.begin() as c:
            c.exec_driver_sql(f'CREATE ROWSTORE TABLE {tmp} (a {coltype})')
        t = con.table(tmp)
        assert t.schema() == ibis.schema({'a': dt.binary})

    finally:
        con.drop_table(tmp, force=True)


@pytest.fixture(scope='session')
def tmp_t(con_nodb: Any) -> Generator[Any, Any, Any]:
    pid = os.getpid()
    with con_nodb.begin() as c:
        c.exec_driver_sql(
            f'CREATE TABLE IF NOT EXISTS {SINGLESTOREDB_DB}.t_{pid} (x TEXT)',
        )
    yield
    with con_nodb.begin() as c:
        c.exec_driver_sql(f'DROP TABLE IF EXISTS {SINGLESTOREDB_DB}.t_{pid}')


@pytest.mark.usefixtures('setup_privs', 'tmp_t')
def test_get_schema_from_query_other_schema(con_nodb: Any) -> None:
    pid = os.getpid()
    t = con_nodb.table(f't_{pid}', schema=f'{SINGLESTOREDB_DB}')
    assert t.schema() == ibis.schema({'x': dt.string})


def test_zero_timestamp_data(con: Any) -> None:
    guid = ibis.util.guid()
    sql = f"""
    CREATE TABLE ztmp_date_issue_{guid}
    (
        name      CHAR(10) NULL,
        tradedate DATETIME NOT NULL,
        date      DATETIME NULL
    );
    """
    try:
        with con.begin() as c:
            c.exec_driver_sql(sql)
            c.exec_driver_sql(
                f"""
                INSERT INTO ztmp_date_issue_{guid} VALUES
                    ('C', '2018-10-22', 0),
                    ('B', '2017-06-07', 0),
                    ('A', '2022-12-21', 0)
                """,
            )
        t = con.table(f'ztmp_date_issue_{guid}')
        result = t.execute().sort_values('name').reset_index(drop=True)
        expected = pd.DataFrame(
            {
                'name': ['C', 'B', 'A'],
                'tradedate': pd.to_datetime(
                    [date(2018, 10, 22), date(2017, 6, 7), date(2022, 12, 21)],
                ),
                'date': [pd.NaT, pd.NaT, pd.NaT],
            },
        ).sort_values('name').reset_index(drop=True)
        tm.assert_frame_equal(result, expected)

    finally:
        con.drop_table(f'ztmp_date_issue_{guid}', force=True)


@pytest.fixture(scope='module')
def enum_t(con: Any) -> Generator[Any, Any, Any]:
    name = gen_name('_enum_')
    t = sa.Table(
        name, sa.MetaData(), sa.Column(
            'sml',
            sqlalchemy_singlestoredb.ENUM('small', 'medium', 'large'),
        ),
    )
    with con.begin() as bind:
        t.create(bind=bind)
        bind.execute(t.insert().values(sml='small'))

    yield con.table(name)
    con.drop_table(name, force=True)


@pytest.mark.parametrize(
    ('expr_fn', 'expected'),
    [
        (methodcaller('startswith', 's'), pd.Series([True], name='sml')),
        (methodcaller('endswith', 'm'), pd.Series([False], name='sml')),
        (methodcaller('re_search', 'mall'), pd.Series([True], name='sml')),
        (methodcaller('lstrip'), pd.Series(['small'], name='sml')),
        (methodcaller('rstrip'), pd.Series(['small'], name='sml')),
        (methodcaller('strip'), pd.Series(['small'], name='sml')),
    ],
    ids=['startswith', 'endswith', 're_search', 'lstrip', 'rstrip', 'strip'],
)
def test_enum_as_string(enum_t: Any, expr_fn: Any, expected: Any) -> None:
    expr = expr_fn(enum_t.sml).name('sml')
    res = expr.execute()
    tm.assert_series_equal(res, expected)



#!/usr/bin/env python
# type: ignore
"""SingleStoreDB connection tests."""
from __future__ import annotations

import os
from pathlib import Path
from typing import Generator
from typing import Iterable

import ibis
import pytest
from ibis.backends.tests.base import RoundHalfToEven
from ibis.backends.tests.base import ServiceBackendTest

ibis.options.interactive = False
ibis.options.sql.default_limit = None
ibis.options.verbose = False

SINGLESTOREDB_ROOT_USER = os.environ.get('IBIS_TEST_SINGLESTOREDB_ROOT_USER', 'root')
SINGLESTOREDB_ROOT_PASSWORD = os.environ.get('IBIS_TEST_SINGLESTOREDB_ROOT_PASSWORD', '')
SINGLESTOREDB_USER = os.environ.get('IBIS_TEST_SINGLESTOREDB_USER', 'ibis')
SINGLESTOREDB_PASSWORD = os.environ.get('IBIS_TEST_SINGLESTOREDB_PASSWORD', 'ibis')
SINGLESTOREDB_DRIVER = os.environ.get('IBIS_TEST_SINGLESTOREDB_DRIVER', 'mysql')
SINGLESTOREDB_HOST = os.environ.get('IBIS_TEST_SINGLESTOREDB_HOST', 'localhost')
SINGLESTOREDB_PORT = int(os.environ.get('IBIS_TEST_SINGLESTOREDB_PORT', 9306))
SINGLESTOREDB_DB = os.environ.get('IBIS_TEST_SINGLESTOREDB_DATABASE', 'ibis_testing')

os.environ['SINGLESTOREDB_URL'] = \
    f'singlestoredb+{SINGLESTOREDB_DRIVER}://' \
    f'{SINGLESTOREDB_USER}:{SINGLESTOREDB_PASSWORD}' \
    f'@{SINGLESTOREDB_HOST}:{SINGLESTOREDB_PORT}/{SINGLESTOREDB_DB}'


class TestConf(ServiceBackendTest, RoundHalfToEven):
    """
    SingleStoreDB connection tests.

    Parameters
    ----------
    data_directory : Path
        Path to input data

    """

    # singlestoredb has the same rounding behavior as postgres
    check_dtype = False
    returned_timestamp_unit = 's'
    supports_window_operations = True
    supports_arrays = False
    supports_structs = False
    supports_arrays_outside_of_select = supports_arrays
    bool_is_int = True

    service_name = 'singlestoredb'
    deps = 'singlestoredb', 'sqlalchemy-singlestoredb', 'sqlalchemy'

    def test_files(self) -> Iterable[Path]:
        return self.data_dir.joinpath('csv').glob('*.csv')

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        database = kw.pop('database', SINGLESTOREDB_DB)
        return ibis.singlestoredb.connect(
            host=SINGLESTOREDB_HOST,
            user=SINGLESTOREDB_USER,
            password=SINGLESTOREDB_PASSWORD,
            driver=SINGLESTOREDB_DRIVER,
            database=database,
            port=SINGLESTOREDB_PORT,
            **kw,
        )


def _random_identifier(suffix):
    return f'__ibis_test_{suffix}_{ibis.util.guid()}'


@pytest.fixture(scope='session')
def setup_privs():
    #    engine = sa.create_engine(
    #        f'singlestoredb://{SINGLESTOREDB_ROOT_USER}:{SINGLESTOREDB_ROOT_PASSWORD}'
    #        f'@{SINGLESTOREDB_HOST}:{SINGLESTOREDB_PORT}',
    #    )
    #    with engine.begin() as con:
    #        # allow the ibis user to use any database
    #        con.exec_driver_sql(f'CREATE DATABASE IF NOT EXISTS `{SINGLESTOREDB_DB}`')
    #        con.exec_driver_sql(
    #            f'GRANT CREATE,SELECT,DROP ON `{SINGLESTOREDB_DB}`.* '
    #            f'TO `{SINGLESTOREDB_USER}`@`%%`',
    #        )
    yield
#    with engine.begin() as con:
#        con.exec_driver_sql(f'DROP DATABASE IF EXISTS `{SINGLESTOREDB_DB}`')


@pytest.fixture(scope='session')
def con():
    # sql_file = os.path.join(os.path.dirname(__file__), 'test.sql')
    # dbname, dbexisted = utils.load_sql(sql_file)

    yield TestConf.connect(tmpdir=None, worker_id=None)

    # if not dbexisted:
    #     utils.drop_database(dbname)


@pytest.fixture(scope='session')
def con_nodb():
    return TestConf.connect(tmpdir=None, worker_id=None, database=None)


@pytest.fixture(scope='module')
def db(con):
    return con.database()


@pytest.fixture(scope='module')
def alltypes(db):
    return db.functional_alltypes


@pytest.fixture(scope='module')
def geotable(con):
    return con.table('geo')


@pytest.fixture(scope='module')
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope='module')
def gdf(geotable):
    return geotable.execute()


@pytest.fixture(scope='module')
def at(alltypes):
    return alltypes.op().sqla_table


@pytest.fixture(scope='module')
def intervals(con):
    return con.table('intervals')


@pytest.fixture
def temp_table(con) -> Generator[str, None, None]:
    """
    Return a temporary table name.

    Parameters
    ----------
    con : ibis.singlestoredb.SingleStoreDBClient

    Yields
    ------
    name : string
        Random table name for a temporary usage.

    """
    name = _random_identifier('table')
    try:
        yield name
    finally:
        con.drop_table(name, force=True)



from __future__ import annotations

from typing import Optional

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class BitCount(Value):
    """Number of bits required to construct a number."""

    arg = rlz.strict_numeric

    output_dtype = dt.int
    output_shape = rlz.shape_like('arg')


def bit_count(arg: ir.NumericValue) -> ir.IntegerValue:
    """
    Return the number of ones in the binary representation of a given number.

    Parameters
    ----------
    arg : int or float
        The number to compute the ones in

    Returns
    -------
    IntegerValue

    """
    return BitCount(arg).to_expr()


ir.NumericValue.bit_count = bit_count


class Conv(Value):
    """Numeric base converter."""

    arg = rlz.one_of([rlz.integer, rlz.string])
    from_base = rlz.integer
    to_base = rlz.integer

    output_dtype = dt.string
    output_shape = rlz.shape_like('arg')


def conv(
    arg: ir.NumericValue, from_base: ir.IntegerValue, to_base: ir.IntegerValue,
) -> ir.IntegerValue:
    """
    Convert number between different bases.

    Parameters
    ----------
    arg : int
        The number to convert
    from_base : int
        Base of input value
    to_base : int
        Base of output value

    Returns
    -------
    IntegerValue

    """
    return Conv(arg, from_base, to_base).to_expr()


ir.NumericValue.conv = conv


class Sigmoid(Value):
    """Sigmoid function."""

    arg = rlz.strict_numeric

    output_dtype = dt.double
    output_shape = rlz.shape_like('arg')


def sigmoid(arg: ir.NumericValue) -> ir.FloatingValue:
    """
    Compute the sigmoid function of `arg`.

    Parameters
    ----------
    arg : int or float
        The number to use

    Returns
    -------
    FloatingValue

    """
    return Sigmoid(arg).to_expr()


ir.NumericValue.sigmoid = sigmoid


class ToNumber(Value):
    """Decimal parser."""

    arg = rlz.string
    format_string = rlz.optional(rlz.string)

    output_dtype = dt.double
    output_shape = rlz.shape_like('arg')


def to_number(
    arg: ir.StringValue, format_string: Optional[ir.StringValue] = None,
) -> ir.DecimalValue:
    """
    Convert a string to a decimal value.

    Parameters
    ----------
    arg : string
        The string value to convert
    format_string : string, optional
        The format of the input value

    Returns
    -------
    DecimalValue

    """
    return ToNumber(arg, format_string).to_expr()


ir.StringValue.to_number = to_number


class Trunc(Value):
    """Truncate decimals."""

    arg = rlz.strict_numeric
    decimals = rlz.optional(rlz.integer)

    output_dtype = dt.double
    output_shape = rlz.shape_like('arg')


def trunc(
    arg: ir.NumericValue, decimals: Optional[ir.IntegerValue] = None,
) -> ir.NumericValue:
    """
    Truncate the number of decimals to `decimals`.

    Parameters
    ----------
    arg : float
        The float value to truncate
    decimals : int, optional
        The number of requested decimals

    Returns
    -------
    NumericValue

    """
    return Trunc(arg, decimals).to_expr()


ir.NumericValue.trunc = trunc


class Truncate(Value):
    """Truncate decimals."""

    arg = rlz.strict_numeric
    decimals = rlz.integer

    output_dtype = dt.double
    output_shape = rlz.shape_like('arg')


def truncate(arg: ir.NumericValue, decimals: ir.IntegerValue) -> ir.NumericValue:
    """
    Truncate the number of decimals to `decimals`.

    Parameters
    ----------
    arg : float
        The float value to truncate
    decimals : int, optional
        The number of requested decimals

    Returns
    -------
    NumericValue

    """
    return Truncate(arg, decimals).to_expr()


ir.NumericValue.truncate = truncate



from __future__ import annotations

import json
from typing import Any

import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONSetDouble(Value):
    """JSON set double value."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)
    value = rlz.double

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_set_double(
    arg: ir.JSONValue | ir.StringValue,
    *key_value: ir.StringValue | ir.NumericValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Set a double value at the specified key path.

    Parameters
    ----------
    arg : JSON or string
        JSON object
    *key_value : strings or ints and double
        Keys or zero-indexed array positions followed by the double value

    Returns
    -------
    String or JSON Column

    """
    return JSONSetDouble(arg, key_value[:-1], key_value[-1]).to_expr()


ir.JSONValue.json_set_double = json_set_double
ir.StringValue.json_set_double = json_set_double


class JSONSetString(Value):
    """JSON set string value."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)
    value = rlz.string

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_set_string(
    arg: ir.JSONValue | ir.StringValue,
    *key_value: ir.StringValue | ir.NumericValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Set a double value at the specified key path.

    Parameters
    ----------
    arg : JSON or string
        JSON object
    *key_value : strings or ints
        Keys or zero-indexed array positions followed by the string value

    Returns
    -------
    String or JSON Column

    """
    return JSONSetString(arg, key_value[:-1], key_value[-1]).to_expr()


ir.JSONValue.json_set_string = json_set_string
ir.StringValue.json_set_string = json_set_string


class JSONSetJSON(Value):
    """JSON set JSON value."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)
    value = rlz.string

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_set_json(
    arg: ir.JSONValue | ir.StringValue,
    *key_value: ir.StringValue | ir.NumericValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Set a JSON value at the specified key path.

    Parameters
    ----------
    arg : JSON or string
        JSON object
    *key_value : strings or ints
        Keys or zero-indexed array positions followed by the JSON value

    Returns
    -------
    String or JSON Column

    """
    return JSONSetJSON(arg, key_value[:-1], key_value[-1]).to_expr()


ir.JSONValue.json_set_json = json_set_json
ir.StringValue.json_set_json = json_set_json


class JSONSet(Value):
    """JSON set any value."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)
    value = rlz.any

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_set(
    arg: ir.JSONValue | ir.StringValue,
    *key_value: ir.StringValue | ir.NumericValue | Any,
) -> ir.JSONValue | ir.StringValue:
    """
    Set any value at the specified key path.

    Parameters
    ----------
    arg : Any
        Any object that can be persisted to a JSON string
    *key_value : strings or ints
        Keys or zero-indexed array positions followed by the JSON value

    Returns
    -------
    String or JSON Column

    """
    if not isinstance(key_value[-1], ir.Value):
        value = json.dumps(key_value[-1])
    return JSONSetJSON(arg, key_value[:-1], value).to_expr()


ir.JSONValue.json_set = json_set
ir.StringValue.json_set = json_set



from __future__ import annotations

from sqlglot.dialects.dialect import Dialect
from sqlglot.dialects.mysql import MySQL


# Define a sqlglot dialect based on MySQL if a real one doesn't exist
if Dialect.get('singlestoredb') is None:

    class SingleStoreDB(MySQL):
        pass


dialect = Dialect['singlestoredb']



from __future__ import annotations

from typing import Optional

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class VectorSort(Value):
    """32-bit float vector sort."""

    arg = binary
    direction = rlz.optional(rlz.string)

    output_dtype = dt.binary
    output_shape = rlz.shape_like('arg')


def vector_sort(
    arg: ir.BinaryValue, direction: Optional[ir.StringValue] = 'asc',
) -> ir.BinaryValue:
    """
    Sort the elements in a 32-bit float vector.

    Parameters
    ----------
    arg : blob
        Vector expression
    direction : str, optional
        Direction of the sort: 'asc' or 'desc'

    Returns
    -------
    Blob column

    """
    if direction is ibis.desc:
        direction = 'desc'
    elif direction is ibis.asc:
        direction = 'asc'
    return VectorSort(arg, direction).to_expr()


ir.BinaryValue.vector_sort = vector_sort


class VectorSortI8(VectorSort):
    """8-bit integer vector sort."""


def vector_sort_i8(
    arg: ir.BinaryValue, direction: Optional[ir.StringValue] = 'asc',
) -> ir.BinaryValue:
    """
    Sort the elements in an 8-bit int vector.

    Parameters
    ----------
    arg : blob
        Vector expression
    direction : str, optional
        Direction of the sort: 'asc' or 'desc'

    Returns
    -------
    Blob column

    """
    if direction is ibis.desc:
        direction = 'desc'
    elif direction is ibis.asc:
        direction = 'asc'
    return VectorSortI8(arg, direction).to_expr()


ir.BinaryValue.vector_sort_i8 = vector_sort_i8


class VectorSortI16(VectorSort):
    """16-bit integer vector sort."""


def vector_sort_i16(
    arg: ir.BinaryValue, direction: Optional[ir.StringValue] = 'asc',
) -> ir.BinaryValue:
    """
    Sort the elements in a 16-bit int vector.

    Parameters
    ----------
    arg : blob
        Vector expression
    direction : str, optional
        Direction of the sort: 'asc' or 'desc'

    Returns
    -------
    Blob column

    """
    if direction is ibis.desc:
        direction = 'desc'
    elif direction is ibis.asc:
        direction = 'asc'
    return VectorSortI16(arg, direction).to_expr()


ir.BinaryValue.vector_sort_i16 = vector_sort_i16


class VectorSortI32(VectorSort):
    """32-bit integer vector sort."""


def vector_sort_i32(
    arg: ir.BinaryValue, direction: Optional[ir.StringValue] = 'asc',
) -> ir.BinaryValue:
    """
    Sort the elements in a 32-bit int vector.

    Parameters
    ----------
    arg : blob
        Vector expression
    direction : str, optional
        Direction of the sort: 'asc' or 'desc'

    Returns
    -------
    Blob column

    """
    if direction is ibis.desc:
        direction = 'desc'
    elif direction is ibis.asc:
        direction = 'asc'
    return VectorSortI32(arg, direction).to_expr()


ir.BinaryValue.vector_sort_i32 = vector_sort_i32


class VectorSortI64(VectorSort):
    """64-bit integer vector sort."""


def vector_sort_i64(
    arg: ir.BinaryValue, direction: Optional[ir.StringValue] = 'asc',
) -> ir.BinaryValue:
    """
    Sort the elements in a 64-bit int vector.

    Parameters
    ----------
    arg : blob
        Vector expression
    direction : str, optional
        Direction of the sort: 'asc' or 'desc'

    Returns
    -------
    Blob column

    """
    if direction is ibis.desc:
        direction = 'desc'
    elif direction is ibis.asc:
        direction = 'asc'
    return VectorSortI64(arg, direction).to_expr()


ir.BinaryValue.vector_sort_i64 = vector_sort_i64


class VectorSortF32(VectorSort):
    """32-bit float vector sort."""


def vector_sort_f32(
    arg: ir.BinaryValue, direction: Optional[ir.StringValue] = 'asc',
) -> ir.BinaryValue:
    """
    Sort the elements in a 32-bit float vector.

    Parameters
    ----------
    arg : blob
        Vector expression
    direction : str, optional
        Direction of the sort: 'asc' or 'desc'

    Returns
    -------
    Blob column

    """
    if direction is ibis.desc:
        direction = 'desc'
    elif direction is ibis.asc:
        direction = 'asc'
    return VectorSortF32(arg, direction).to_expr()


ir.BinaryValue.vector_sort_f32 = vector_sort_f32


class VectorSortF64(VectorSort):
    """64-bit float vector sort."""


def vector_sort_f64(
    arg: ir.BinaryValue, direction: Optional[ir.StringValue] = 'asc',
) -> ir.BinaryValue:
    """
    Sort the elements in a 64-bit float vector.

    Parameters
    ----------
    arg : blob
        Vector expression
    direction : str, optional
        Direction of the sort: 'asc' or 'desc'

    Returns
    -------
    Blob column

    """
    if direction is ibis.desc:
        direction = 'desc'
    elif direction is ibis.asc:
        direction = 'asc'
    return VectorSortF64(arg, direction).to_expr()


ir.BinaryValue.vector_sort_f64 = vector_sort_f64



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONExtractDouble(Value):
    """JSON extract double value."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)

    output_dtype = dt.double
    output_shape = rlz.shape_like('arg')


def json_extract_double(
    arg: ir.JSONValue | ir.StringValue,
    *key_path: ir.StringValue | ir.NumericValue,
) -> ir.FloatingValue:
    """
    Extract a double value from a JSON object.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    *key_path : strings or ints
        Keys or zero-indexed array positions.

    Returns
    -------
    Double

    """
    return JSONExtractDouble(arg, key_path).to_expr()


ir.JSONValue.json_extract_double = json_extract_double
ir.StringValue.json_extract_double = json_extract_double


class JSONExtractString(Value):
    """JSON extract string value."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)

    output_dtype = dt.string
    output_shape = rlz.shape_like('arg')


def json_extract_string(
    arg: ir.JSONValue | ir.StringValue,
    *key_path: ir.StringValue | ir.NumericValue,
) -> ir.StringValue:
    """
    Extract a string value from a JSON object.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    *key_path : strings or ints
        Keys or zero-indexed array positions.

    Returns
    -------
    String

    """
    return JSONExtractString(arg, key_path).to_expr()


ir.JSONValue.json_extract_string = json_extract_string
ir.StringValue.json_extract_string = json_extract_string


class JSONExtractJSON(Value):
    """JSON extract string value."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_extract_json(
    arg: ir.JSONValue | ir.StringValue,
    *key_path: ir.StringValue | ir.NumericValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Extract a JSON value from a JSON object.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    *key_path : strings or ints
        Keys or zero-indexed array positions.

    Returns
    -------
    JSON or string

    """
    return JSONExtractJSON(arg, key_path).to_expr()


ir.JSONValue.json_extract_json = json_extract_json
ir.StringValue.json_extract_json = json_extract_json


class JSONExtractBigint(Value):
    """JSON extract integer value."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)

    output_dtype = dt.int64
    output_shape = rlz.shape_like('arg')


def json_extract_bigint(
    arg: ir.JSONValue | ir.StringValue,
    *key_path: ir.StringValue | ir.NumericValue,
) -> ir.IntegerValue:
    """
    Extract an int value from a JSON object.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    *key_path : strings or ints
        Keys or zero-indexed array positions.

    Returns
    -------
    Integer

    """
    return JSONExtractBigint(arg, key_path).to_expr()


ir.JSONValue.json_extract_bigint = json_extract_bigint
ir.StringValue.json_extract_bigint = json_extract_bigint



#!/usr/bin/env python3
# type: ignore
from __future__ import annotations

import ibis
import ibis.expr.datatypes as dt
import pytest
from pytest import param

ibis.options.verbose = False

OPERATIONS = [
    # Abs
    ('abs()', lambda x: x.abs()),
    # Add
    ('add()', lambda x: x.add(3)),
    # Capitalize
    ('capitalize()', lambda x: x.capitalize()),
    # Case
    ('cast()', lambda x: x.cast(dt.string)),
    # Ceil
    ('ceil()', lambda x: x.ceil()),
    # Bucket
    ('bucket()', lambda x: x.bucket([0, 50, 100])),
    # Coalesce
    ('coalesce()', lambda x: x.coalesce()),
    # And
    # BitAnd
    ('bit_and()', lambda x: x.bit_and(True)),
    # BitOr
    ('bit_or()', lambda x: x.bit_or(True)),
    # BitXor
    ('bit_xor()', lambda x: x.bit_xor(True)),
    # CategoryLabel
    # CountDistinct
    ('count_distinct()', lambda x: x.count()),
    # Count
    ('count()', lambda x: x.count()),
    # CumulativeMax
    ('cummax()', lambda x: x.cummax()),
    # CumulativeMean
    ('cummean()', lambda x: x.cummean()),
    # CumulativeMin
    ('cummin()', lambda x: x.cummin()),
    # CumulativeSum
    ('cumsum()', lambda x: x.cumsum()),
    # DateAdd
    # DateDiff
    # DateFromYMD
    # DateSub
    # DateTruncate
    # Date
    # DayOfWeekIndex
    # DayOfWeekName
    # DenseRank
    ('dense_rank()', lambda x: x.dense_rank()),
    # EndsWith
    ('endswith()', lambda x: x.endswith('a')),
    # Equals
    # ExistsSubquery
    # Exp
    # ExtractDayOfYear
    # ExtractDay
    # ExtractEpochSeconds
    # ExtractHour
    # ExtractMillisecond
    # ExtractMinute
    # ExtractMonth
    # ExtractQuarter
    # ExtractSecond
    # ExtractWeekOfYear
    # ExtractYear
    # FirstValue
    # ('first()', lambda x: x.first()),
    # FloorDivide
    # Floor
    ('floor()', lambda x: x.floor()),
    # GeoArea
    # GeoAsBinary
    # GeoAsEWKB
    # GeoAsEWKT
    # GeoAsText
    # GeoAzimuth
    # GeoBuffer
    # GeoCentroid
    # GeoContainsProperly
    # GeoContains
    # GeoCoveredBy
    # GeoCovers
    # GeoCrosses
    # GeoDFullyWithin
    # GeoDWithin
    # GeoDifference
    # GeoDisjoint
    # GeoDistance
    # GeoEndPoint
    # GeoEnvelope
    # GeoEquals
    # GeoGeometryN
    # GeoGeometryType
    # GeoIntersection
    # GeoIntersects
    # GeoIsValid
    # GeoLength
    # GeoLineLocatePoint
    # GeoLineMerge
    # GeoLineSubstring
    # GeoNPoints
    # GeoOrderingEquals
    # GeoOverlaps
    # GeoPerimeter
    # GeoSRID
    # GeoSetSRID
    # GeoSimplify
    # GeoStartPoint
    # GeoTouches
    # GeoTransform
    # GeoUnaryUnion
    # GeoUnion
    # GeoWithin
    # GeoX
    # GeoY
    # GreaterEqual
    # Greater
    # Greatest
    ('greatest()', lambda x: x.greatest(x.min())),
    # GroupConcat
    ('group_concat()', lambda x: x.group_concat()),
    # IdenticalTo
    # IfNull
    # IntervalFromInteger
    # IsNull
    # LPad
    # LStrip
    # Lag
    # LastValue
    # Lead
    ('lead()', lambda x: x.lead()),
    # Least
    ('least()', lambda x: x.least(x.max())),
    ('cumsum()', lambda x: x.cumsum()),
    # LessEqual
    # Less
    # Ln
    ('ln()', lambda x: x.ln()),
    # Log10
    ('log10()', lambda x: x.log10()),
    # Log2
    ('log2()', lambda x: x.log2()),
    # Log
    ('log()', lambda x: x.log(11)),
    # Lowercase
    # Max
    ('max()', lambda x: x.max()),
    # Mean
    ('mean()', lambda x: x.mean()),
    # MinRank
    # Min
    ('min()', lambda x: x.min()),
    # Modulus
    ('mod()', lambda x: x.mod(8)),
    # Multiply
    ('mul()', lambda x: x.mul(3)),
    # NTile
    # ('ntile()', lambda x: x.ntile(2)),
    # Negate
    ('negate()', lambda x: x.negate()),
    # NotAll
    # NotAny
    # NotContains
    # NotEquals
    # NotNull
    ('notnull()', lambda x: x.notnull()),
    # Not
    # NullIfZero
    ('nullifzero()', lambda x: x.nullifzero()),
    # NullIf
    ('nullif()', lambda x: x.nullif(x.max())),
    # NullLiteral
    # Or
    # PercentRank
    ('percent_rank()', lambda x: x.percent_rank()),
    # Power
    # RPad
    # RStrip
    # RandomScalar
    # RegexSearch
    # Repeat
    # Reverse
    # Round
    ('round()', lambda x: x.round(4)),
    # RowNumber
    # SearchedCase
    # Sign
    ('sign()', lambda x: x.sign()),
    # SimpleCase
    # Sqrt
    ('sqrt()', lambda x: x.sqrt()),
    # StandardDev
    ('std()', lambda x: x.std()),
    # StartsWith
    # StrRight
    # Strftime
    # StringAscii
    # StringConcat
    # StringContains
    # StringFind
    # StringJoin
    # StringLength
    # StringReplace
    # StringSQLILike
    # Strip
    # Substring
    # Subtract
    # Sum
    ('sum()', lambda x: x.sum()),
    # TableArrayView
    # TableColumn
    # TimeFromHMS
    # TimestampAdd
    # TimestampDiff
    # TimestampFromYMDHMS
    # TimestampNow
    # TimestampSub
    # TimestampTruncate
    # TypeOf
    # Uppercase
    # ValueList
    # Variance
    # WindowOp
]

COLUMNS = [
    'tinyint_c',
    'smallint_c',
    'mediumint_c',
    'int_c',
    'bigint_c',
    'float_c',
    'double_c',
    'decimal_c',
    'decimal5_c',
    'numeric_c',
    'date_c',
    'time_c',
    'time6_c',
    'datetime_c',
    'datetime6_c',
    'timestamp_c',
    'timestamp6_c',
    'char32_c',
    'varchar42_c',
    'longtext_c',
    'mediumtext_c',
    'tinytext_c',
    'text_c',
    'text4_c',
    'blob_c',
    'enum_sml_c',
    'set_abcd_c',
    'negative_int_c',
    'negative_float_c',
    'bool_c',
]


@pytest.mark.parametrize(
    ('operation', 'column', 'ibis_operation'),
    [
        param(operation, column, ibis_operation, id=operation)
        for operation, ibis_operation in OPERATIONS
        for column in COLUMNS
    ],
)
def test_operations(con, operation, column, ibis_operation):
    datatypes = con.table('datatypes')

    try:
        ibis_operation(getattr(datatypes, column)).name('_').execute()
    except AttributeError:
        return

    print(operation, 'is supported for', column)



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class VectorElementsSum(Value):
    """32-bit float JSON array unpacker."""

    arg = binary

    output_dtype = dt.float64
    output_shape = rlz.shape_like('arg')


def vector_elements_sum(arg: ir.BinaryValue) -> ir.FloatingValue:
    """
    Convert an encoded blob representing a 32-bit float vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return VectorElementsSum(arg).to_expr()


ir.BinaryValue.vector_elements_sum = vector_elements_sum


class VectorElementsSumI8(VectorElementsSum):
    """8-bit integer JSON array unpacker."""

    output_dtype = dt.int64


def vector_elements_sum_i8(arg: ir.BinaryValue) -> ir.IntegerValue:
    """
    Convert an encoded blob representing an 8-bit int vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return VectorElementsSumI8(arg).to_expr()


ir.BinaryValue.vector_elements_sum_i8 = vector_elements_sum_i8


class VectorElementsSumI16(VectorElementsSum):
    """16-bit integer JSON array unpacker."""

    output_dtype = dt.int64


def vector_elements_sum_i16(arg: ir.BinaryValue) -> ir.IntegerValue:
    """
    Convert an encoded blob representing a 16-bit int vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return VectorElementsSumI16(arg).to_expr()


ir.BinaryValue.vector_elements_sum_i16 = vector_elements_sum_i16


class VectorElementsSumI32(VectorElementsSum):
    """32-bit integer JSON array unpacker."""

    output_dtype = dt.int64


def vector_elements_sum_i32(arg: ir.BinaryValue) -> ir.IntegerValue:
    """
    Convert an encoded blob representing a 32-bit int vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return VectorElementsSumI32(arg).to_expr()


ir.BinaryValue.vector_elements_sum_i32 = vector_elements_sum_i32


class VectorElementsSumI64(VectorElementsSum):
    """64-bit integer JSON array unpacker."""

    output_dtype = dt.int64


def vector_elements_sum_i64(arg: ir.BinaryValue) -> ir.IntegerValue:
    """
    Convert an encoded blob representing a 64-bit int vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return VectorElementsSumI64(arg).to_expr()


ir.BinaryValue.vector_elements_sum_i64 = vector_elements_sum_i64


class VectorElementsSumF32(VectorElementsSum):
    """32-bit float JSON array unpacker."""

    output_dtype = dt.double


def vector_elements_sum_f32(arg: ir.BinaryValue) -> ir.FloatingValue:
    """
    Convert an encoded blob representing a 32-bit float vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return VectorElementsSumF32(arg).to_expr()


ir.BinaryValue.vector_elements_sum_f32 = vector_elements_sum_f32


class VectorElementsSumF64(VectorElementsSum):
    """64-bit float JSON array unpacker."""

    output_dtype = dt.double


def vector_elements_sum_f64(arg: ir.BinaryValue) -> ir.FloatingValue:
    """
    Convert an encoded blob representing a 64-bit float vector to a JSON array.

    Parameters
    ----------
    arg : blob
        Encoded array

    Returns
    -------
    JSON column

    """
    return VectorElementsSumF64(arg).to_expr()


ir.BinaryValue.vector_elements_sum_f64 = vector_elements_sum_f64



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class JSONArrayPack(Value):
    """32-bit float JSON array packer."""

    arg = rlz.one_of([rlz.string, rlz.json])

    output_dtype = dt.binary
    output_shape = rlz.shape_like('arg')


def json_array_pack(arg: ir.StringValue | ir.JSONValue) -> ir.BinaryValue:
    """
    Convert a JSON array of zero or more 32-bit floats to an encoded blob.

    Parameters
    ----------
    arg : string
        JSON array

    Returns
    -------
    Blob column

    """
    return JSONArrayPack(arg).to_expr()


ir.StringValue.json_array_pack = json_array_pack
ir.JSONValue.json_array_pack = json_array_pack


class JSONArrayPackI8(JSONArrayPack):
    """8-bit integer JSON array packer."""


def json_array_pack_i8(arg: ir.StringValue | ir.JSONValue) -> ir.BinaryValue:
    """
    Convert a JSON array of zero or more 8-bit ints to an encoded blob.

    Parameters
    ----------
    arg : string
        JSON array

    Returns
    -------
    Blob column

    """
    return JSONArrayPackI8(arg).to_expr()


ir.StringValue.json_array_pack_i8 = json_array_pack_i8
ir.JSONValue.json_array_pack_i8 = json_array_pack_i8


class JSONArrayPackI16(JSONArrayPack):
    """16-bit integer JSON array packer."""


def json_array_pack_i16(arg: ir.StringValue | ir.JSONValue) -> ir.BinaryValue:
    """
    Convert a JSON array of zero or more 16-bit ints to an encoded blob.

    Parameters
    ----------
    arg : string
        JSON array

    Returns
    -------
    Blob column

    """
    return JSONArrayPackI16(arg).to_expr()


ir.StringValue.json_array_pack_i16 = json_array_pack_i16
ir.JSONValue.json_array_pack_i16 = json_array_pack_i16


class JSONArrayPackI32(JSONArrayPack):
    """32-bit integer JSON array packer."""


def json_array_pack_i32(arg: ir.StringValue | ir.JSONValue) -> ir.BinaryValue:
    """
    Convert a JSON array of zero or more 32-bit ints to an encoded blob.

    Parameters
    ----------
    arg : string
        JSON array

    Returns
    -------
    Blob column

    """
    return JSONArrayPackI32(arg).to_expr()


ir.StringValue.json_array_pack_i32 = json_array_pack_i32
ir.JSONValue.json_array_pack_i32 = json_array_pack_i32


class JSONArrayPackI64(JSONArrayPack):
    """64-bit integer JSON array packer."""


def json_array_pack_i64(arg: ir.StringValue | ir.JSONValue) -> ir.BinaryValue:
    """
    Convert a JSON array of zero or more 64-bit ints to an encoded blob.

    Parameters
    ----------
    arg : string
        JSON array

    Returns
    -------
    Blob column

    """
    return JSONArrayPackI64(arg).to_expr()


ir.StringValue.json_array_pack_i64 = json_array_pack_i64
ir.JSONValue.json_array_pack_i64 = json_array_pack_i64


class JSONArrayPackF32(JSONArrayPack):
    """32-bit float JSON array packer."""


def json_array_pack_f32(arg: ir.StringValue | ir.JSONValue) -> ir.BinaryValue:
    """
    Convert a JSON array of zero or more 32-bit floats to an encoded blob.

    Parameters
    ----------
    arg : string
        JSON array

    Returns
    -------
    Blob column

    """
    return JSONArrayPackF32(arg).to_expr()


ir.StringValue.json_array_pack_f32 = json_array_pack_f32
ir.JSONValue.json_array_pack_f32 = json_array_pack_f32


class JSONArrayPackF64(JSONArrayPack):
    """64-bit float JSON array packer."""


def json_array_pack_f64(arg: ir.StringValue | ir.JSONValue) -> ir.BinaryValue:
    """
    Convert a JSON array of zero or more 64-bit floats to an encoded blob.

    Parameters
    ----------
    arg : string
        JSON array

    Returns
    -------
    Blob column

    """
    return JSONArrayPackF64(arg).to_expr()


ir.StringValue.json_array_pack_f64 = json_array_pack_f64
ir.JSONValue.json_array_pack_f64 = json_array_pack_f64



from __future__ import annotations

import json
from typing import Any

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONArrayContainsDouble(Value):
    """JSON array element test."""

    arg = rlz.one_of([rlz.json, rlz.string])
    value = rlz.strict_numeric

    output_dtype = dt.bool
    output_shape = rlz.shape_like('arg')


def json_array_contains_double(
    arg: ir.JSONValue | ir.StringValue, value: ir.NumericValue,
) -> ir.BooleanValue:
    """
    Does the array contain the given float value?

    Parameters
    ----------
    arg : JSON or string
        JSON array
    value : float
        Float value to search for

    Returns
    -------
    Boolean column

    """
    return JSONArrayContainsDouble(arg, value).to_expr()


ir.JSONValue.json_array_contains_double = json_array_contains_double
ir.StringValue.json_array_contains_double = json_array_contains_double


class JSONArrayContainsString(Value):
    """JSON array element test."""

    arg = rlz.one_of([rlz.json, rlz.string])
    value = rlz.string

    output_dtype = dt.bool
    output_shape = rlz.shape_like('arg')


def json_array_contains_string(
    arg: ir.JSONValue | ir.StringValue, value: ir.StringValue,
) -> ir.BooleanValue:
    """
    Does the array contain the given string value?

    Parameters
    ----------
    arg : JSON or string
        JSON array
    value : string
        String value to search for

    Returns
    -------
    Boolean column

    """
    return JSONArrayContainsString(arg, value).to_expr()


ir.JSONValue.json_array_contains_string = json_array_contains_string
ir.StringValue.json_array_contains_string = json_array_contains_string


class JSONArrayContainsJSON(Value):
    """JSON array element test."""

    arg = rlz.one_of([rlz.json, rlz.string])
    value = rlz.one_of([rlz.json, rlz.string])

    output_dtype = dt.bool
    output_shape = rlz.shape_like('arg')


def json_array_contains_json(
    arg: ir.JSONValue | ir.StringValue, value: ir.JSONValue | ir.StringValue,
) -> ir.BooleanValue:
    """
    Does the array contain the given JSON value?

    Parameters
    ----------
    arg : JSON or string
        JSON array
    value : string
        JSON value to search for

    Returns
    -------
    Boolean column

    """
    return JSONArrayContainsJSON(arg, value).to_expr()


ir.JSONValue.json_array_contains_json = json_array_contains_json
ir.StringValue.json_array_contains_json = json_array_contains_json


class JSONArrayContains(Value):
    """JSON array element test."""

    arg = rlz.one_of([rlz.json, rlz.string])
    value = rlz.any

    output_dtype = dt.bool
    output_shape = rlz.shape_like('arg')


def json_array_contains(
    arg: ir.JSONValue | ir.StringValue, value: ir.Value | Any,
) -> ir.BooleanValue:
    """
    Does the array contain the given value?

    Parameters
    ----------
    arg : JSON or string
        JSON object
    value : Any
        Value to search for

    Returns
    -------
    Boolean column

    """
    if not isinstance(value, ir.Value):
        value = json.dumps(value)
    return JSONArrayContainsJSON(arg, value).to_expr()


ir.JSONValue.json_array_contains = json_array_contains
ir.StringValue.json_array_contains = json_array_contains



from __future__ import annotations

import json
from typing import Iterable

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class EuclideanDistance(Value):
    """32-bit float Euclidean distance."""

    left = binary
    right = binary

    output_dtype = dt.double
    output_shape = rlz.shape_like('left')


def euclidean_distance(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar Euclidean distance between two 32-bit float input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack()
    return EuclideanDistance(left, right).to_expr()


ir.BinaryValue.euclidean_distance = euclidean_distance


class EuclideanDistanceI8(EuclideanDistance):
    """8-bit integer Euclidean distance."""


def euclidean_distance_i8(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar Euclidean distance between two 8-bit int input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i8()
    return EuclideanDistanceI8(left, right).to_expr()


ir.BinaryValue.euclidean_distance_i8 = euclidean_distance_i8


class EuclideanDistanceI16(EuclideanDistance):
    """16-bit integer Euclidean distance."""


def euclidean_distance_i16(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar Euclidean distance between two 16-bit int input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i16()
    return EuclideanDistanceI16(left, right).to_expr()


ir.BinaryValue.euclidean_distance_i16 = euclidean_distance_i16


class EuclideanDistanceI32(EuclideanDistance):
    """32-bit integer Euclidean distance."""


def euclidean_distance_i32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar Euclidean distance between two 32-bit int input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i32()
    return EuclideanDistanceI32(left, right).to_expr()


ir.BinaryValue.euclidean_distance_i32 = euclidean_distance_i32


class EuclideanDistanceI64(EuclideanDistance):
    """64-bit integer Euclidean distance."""


def euclidean_distance_i64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar Euclidean distance between two 64-bit int input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i64()
    return EuclideanDistanceI64(left, right).to_expr()


ir.BinaryValue.euclidean_distance_i64 = euclidean_distance_i64


class EuclideanDistanceF32(EuclideanDistance):
    """32-bit float Euclidean distance."""


def euclidean_distance_f32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar Euclidean distance between two 32-bit float input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_f32()
    return EuclideanDistanceF32(left, right).to_expr()


ir.BinaryValue.euclidean_distance_f32 = euclidean_distance_f32


class EuclideanDistanceF64(EuclideanDistance):
    """64-bit float Euclidean distance."""


def euclidean_distance_f64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar Euclidean distance between two 64-bit float input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_f64()
    return EuclideanDistanceF64(left, right).to_expr()


ir.BinaryValue.euclidean_distance_f64 = euclidean_distance_f64



#!/usr/bin/env python3
from __future__ import annotations

import re
import warnings
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import ibis
import ibis.backends.base.sql.compiler.translator as tr
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.rules as rlz
import ibis.expr.types as types
import sqlalchemy as sa


class Table(dt.DataType):
    """Placeholder for Table types until Ibis supports them."""


def _build_data_type(
    dtype: str,
    info: Dict[str, Any],
    args: Optional[Sequence[int]] = None,
    schema: Optional[Sequence[dt.DataType]] = None,
) -> dt.DataType:
    """
    Build a dt.DataType from given SingleStoreDB data information.

    Parameters
    ----------
    dtype : str
        Name of the data type
    info : Dict[str, Any]
        Data type modifiers
    args : Sequence[int], optional
        Data type parameters
    schema : Sequence[dt.DataType], optional
        Schema of data type if it is a array, record, or table

    Returns
    -------
    dt.DataType

    """
    type_map = {
        'bool': dt.Boolean,
        'bit': dt.Binary,
        'tinyint': dt.Int8,
        'smallint': dt.Int16,
        'mediumint': dt.Int32,
        'int': dt.Int32,
        'bigint': dt.Int64,
        'float': dt.Float32,
        'double': dt.Float64,
        'tinyint unsigned': dt.UInt8,
        'smallint unsigned': dt.UInt16,
        'mediumint unsigned': dt.UInt32,
        'int unsigned': dt.UInt32,
        'bignt unsigned': dt.UInt64,
        'decimal': dt.Decimal,
        'date': dt.Date,
        'time': dt.Interval,
        'datetime': dt.Timestamp,
        'timestamp': dt.Timestamp,
        'year': dt.Int16,
        'char': dt.String,
        'varchar': dt.String,
        'text': dt.String,
        'tinytext': dt.String,
        'mediumtext': dt.String,
        'longtext': dt.String,
        'binary': dt.Binary,
        'varbinary': dt.Binary,
        'blob': dt.Binary,
        'tinyblob': dt.Binary,
        'mediumblob': dt.Binary,
        'longblob': dt.Binary,
        'json': dt.JSON,
        'record': dt.Struct,
        'geograph': dt.Geography,
        'geographypoint': dt.Point,
        'table': Table,
        'null': dt.Null,
        'array': dt.Array,
    }

    attrs: Dict[str, Any] = {}

    if args and 'decimal' in dtype:
        attrs['precision'] = args[0]
        if len(args) > 1:
            attrs['scale'] = args[1]

    if dtype in ['datetime', 'timestamp']:
        attrs['value_type'] = dt.Int64()

    elif dtype in ['record']:
        if not schema:
            raise ValueError('A schema is required for record and table types.')
        attrs['names'] = [x[0] for x in schema]
        attrs['types'] = [x[1] for x in schema]

    elif dtype in ['array']:
        if not schema:
            raise ValueError('A data type is required for array types.')
        attrs['value_type'] = schema[0][1]

    attrs['nullable'] = info.get('nullable', False)

    return type_map[dtype](**attrs)


def _parse_data_type(data: str) -> Tuple[dt.DataType, str]:
    """
    Parse data type from string.

    Parameters
    ----------
    data : str
        String to parse

    Returns
    -------
    Tuple of parsed data type and remaining string

    """
    _, data_type, data = re.split(
        r'\s*(\w+(?:\s+UNSIGNED)?)\s*',
        data,
        flags=re.I,
        maxsplit=1,
    )
    data_type = data_type.lower()

    schema = None
    data_type_args: List[int] = []

    if data_type in ['table', 'record', 'array']:
        _, data = re.split(r'^\s*\(\s*', data, flags=re.I, maxsplit=1)
        if data_type in ['array']:
            schema, data = _parse_params(data, parse_names=False)
        else:
            schema, data = _parse_params(data)
    elif data.startswith('('):
        _, data = re.split(r'^\s*\(\s*', data, maxsplit=1)
        args, data = re.split(r'\s*\)\s*', data, flags=re.I, maxsplit=1)
        data_type_args = [int(x) for x in re.split(r'\s*,\s*', args)]

    modifiers = {}
    while data and re.match(
        r'^(CHARACTER\s+SET\s+|COLLATE\s+|NULL|NOT\s+NULL)',
        data,
        flags=re.I,
    ):
        if re.match(r'CHARACTER\s+SET\s+', data, flags=re.I):
            _, modifiers['character_set'], data = re.split(
                r'CHARACTER\s+SET\s+(\S+)\s*',
                data,
                flags=re.I,
                maxsplit=1,
            )
        elif re.match(r'COLLATE\s+', data, flags=re.I):
            _, modifiers['collate'], data = re.split(
                r'COLLATE\s+(\S+)\s*',
                data,
                flags=re.I,
                maxsplit=1,
            )
        elif re.match(r'NULL', data, flags=re.I):
            _, data = re.split(r'NULL\s*', data, flags=re.I, maxsplit=1)
            modifiers['nullable'] = True
        elif re.match(r'NOT\s+NULL', data, flags=re.I):
            _, data = re.split(r'NOT\s+NULL\s*', data, flags=re.I, maxsplit=1)
            modifiers['nullable'] = False
        else:
            unknown, data = re.split(r'(\S+)', data, flags=re.I, maxsplit=1)
            warnings.warn(
                f'Skipping unknown CREATE FUNCTION syntax: {unknown}',
                RuntimeWarning,
            )

    return _build_data_type(data_type, modifiers, data_type_args, schema), data


def _parse_params(
    params: str, parse_names: bool = True,
) -> Tuple[List[dt.DataType], str]:
    """
    Parse function / table / record / array parameters from string.

    Parameters
    ----------
    params : str
        String to parse
    parse_names : bool, optional
        Do the parameters include parameter names?

    Returns
    -------
    List of parsed datatypes and remaining string

    """
    out = []
    i = ord('a')
    while params and not params.startswith(')'):
        if parse_names:
            _, param_name, params = re.split(
                r'(\S+)\s+', params, flags=re.I, maxsplit=1,
            )

            if param_name.startswith('`') and param_name.endswith('`'):
                param_name = param_name[1:-1]
        else:
            param_name = chr(i)
            i += 1
            params = params.lstrip()

        param_type, params = _parse_data_type(params)

        if params.startswith(','):
            params = re.sub(r'^,\s*', r'', params)

        out.append((param_name, param_type))

    return out, re.sub(r'^\s*\)\s*', r'', params, flags=re.I)


def _parse_create_function(
    func: str,
) -> Tuple[str, str, List[dt.DataType], dt.DataType, Dict[str, Any]]:
    """
    Parse a CREATE FUNCTION prototype.

    Parameters
    ----------
    func : str
        String to parse

    Returns
    -------
    Tuple of function type, function name, list of data types, return data type,
    and dictionary of function metadata

    """
    # Strip CREATE keyword
    func = re.sub(r'^\s*CREATE\s+(OR\s+REPLACE\s+)?\s*', r'', func, flags=re.I)

    # Get function type
    _, func_type, func = re.split(
        r'((?:EXTERNAL\s+)?(?:FUNCTION|AGGREGATE))\s+',
        func,
        flags=re.I,
        maxsplit=1,
    )
    func_type = func_type.lower()

    # Get function name
    func_name, func = re.split(r'\s*\(\s*', func, flags=re.I, maxsplit=1)
    if func_name.startswith('`') and func_name.endswith('`'):
        func_name = func_name[1:-1]

    # Parse parameters
    inputs, func = _parse_params(func, parse_names='agg' not in func_type)

    # Bypass RETURNS keyword
    func = re.sub(r'^\s*RETURNS\s+', r'', func, flags=re.I)

    # Parse return type
    return_type, func = _parse_data_type(func)

    # Detect implementation-specific information
    info: Dict[str, Any] = {}

    m_wasm = re.search(r'\bAS\s+WASM\b', func, flags=re.I)
    if m_wasm:
        info['wasm'] = True

    m_remote_service = re.search(
        r'\bAS\s+REMOTE\s+SERVICE\s*("|\')([^\1])\1',
        func,
        flags=re.I,
    )
    if m_remote_service:
        info['remote_service'] = m_remote_service.group(2)

    m_format = re.search(r'\bFORMAT\s+(\S+)', func, flags=re.I)
    if m_format:
        info['format'] = m_format.group(1).lower()

    return func_type, func_name, inputs, return_type, info


def build_function(conn: Any, name: str) -> Optional[Callable[..., Any]]:
    """
    Build a Ibis function from a given function name.

    Parameters
    ----------
    conn : Ibis Connection
        Ibis connection object
    name : str
        Name of the function to build

    Returns
    -------
    Callable Ibis function

    """
    quote_identifier = conn.con.dialect.identifier_preparer.quote_identifier
    db = quote_identifier(conn.database_name)
    qname = quote_identifier(name)
    proto = conn.raw_sql(f'show create function {db}.{qname}').fetchall()[0][2]

    try:
        func_type, func_name, inputs, output, info = _parse_create_function(proto)
    except Exception:
        print(f'Failed to parse: {proto}')
        return None

    return _make_udf(func_name, func_type, inputs, output, info)


def _make_func_doc(
    name: str,
    ftype: str,
    inputs: Sequence[Tuple[str, dt.DataType]],
    output: Optional[dt.DataType],
    info: Optional[Dict[str, Any]],
) -> str:
    """
    Construct the docstring using the function information.

    Parameters
    ----------
    name : str
        Name of the function
    ftype : str
        Type of the function in the database
    inputs : Sequence[Tuple[str, dt.DataType]]
        Sequence of (name, type) elements describing
        the inputs of the function
    output : dt.DataType, optional
        Data type of the return value of the function
    info : Dict[str, Any], optional
        Function metadata

    Returns
    -------
    str

    """
    info = info or {}
    code = info.get('code', '')
    format = info.get('format', '')
    doc = [f'Call `{name}` {ftype} function.', '']
    if ftype == 'remote service':
        doc.append(f'Accesses remote service at {code} using {format} format.')
        doc.append('')
    doc.extend(['Parameters', '----------'])
    for name, dtype in inputs:
        arg = f'{name} : {dtype}'
        if dtype.nullable:
            arg += ' or None'
        doc.append(arg)
    if output is not None:
        doc.append('')
        doc.extend(['Returns', '-------'])
        ret = str(output)
        if output.nullable:
            ret += ' or None'
        doc.append(ret)
    doc.append('')
    return '\n'.join(doc)


def _make_udf(
    name: str,
    ftype: str,
    inputs: Sequence[Tuple[str, dt.DataType]],
    output: Optional[dt.DataType],
    info: Optional[Dict[str, Any]],
) -> Optional[Callable[..., Any]]:
    """Define a callable that invokes a UDF on the server."""
    if output is not None and isinstance(output, Table):
        warnings.warn(
            f'Could not create function `{name}`. '
            'Table return types are not supported.',
            RuntimeWarning,
        )
        return None

    info = info or {}

    # Define generic function API
    cls_params = {}

    for arg, dtype in inputs:
        cls_params[arg] = rlz.value(dtype)

    if output:
        cls_params['output_dtype'] = output
        # TODO: Verify shape for various output types
        if inputs:
            cls_params['output_shape'] = rlz.shape_like(inputs[0][0])

    func_type = type(name.title().replace('_', ''), (ops.ValueOp,), cls_params)

    def eval_func(*args: Any) -> types.Expr:
        return func_type(*args).to_expr()

    eval_func.__name__ = name
    eval_func.__qualname__ = f'ibis.singlestoredb.{name}'
    eval_func.__doc__ = _make_func_doc(name, ftype, inputs, output, info)

    # TODO: Check for existing function
    if inputs:
        if isinstance(inputs[0][1], dt.String):
            setattr(types.StringValue, name, eval_func)
        elif isinstance(inputs[0][1], dt.Integer):
            setattr(types.IntegerValue, name, eval_func)
        elif isinstance(inputs[0][1], dt.Floating):
            setattr(types.FloatingValue, name, eval_func)

    @ibis.singlestoredb.add_operation(func_type)
    def _eval_func(t: tr.ExprTranslator, op: ops.ValueOp) -> types.Expr:
        return getattr(sa.func, name)(*[t.translate(x) for x in op.args])

    return eval_func



from __future__ import annotations

import contextlib
import functools
import operator
import warnings
from typing import Any
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Union

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
import ibis.expr.types.groupby as gby
import numpy as np
import pandas as pd
import sqlalchemy as sa
from ibis import util
from ibis.backends.base.sql.alchemy import AlchemyExprTranslator as ExprTranslator
from ibis.backends.base.sql.alchemy import fixed_arity
from ibis.backends.base.sql.alchemy import sqlalchemy_operation_registry
from ibis.backends.base.sql.alchemy import sqlalchemy_window_functions_registry
from ibis.backends.base.sql.alchemy import unary
from ibis.backends.base.sql.alchemy.geospatial import geospatial_supported
from ibis.backends.base.sql.alchemy.registry import geospatial_functions
from ibis.backends.base.sql.alchemy.registry import reduction

from . import functions


operation_registry = sqlalchemy_operation_registry.copy()

# NOTE: window functions are available from MySQL 8 and MariaDB 10.2
operation_registry.update(sqlalchemy_window_functions_registry)

if geospatial_supported:
    operation_registry.update(geospatial_functions)


def _substr(t: ExprTranslator, op: ops.Substring) -> ir.Expr:
    f = sa.func.substr

    sa_arg = t.translate(op.arg)
    sa_start = t.translate(op.start)

    if op.length is None:
        return sa.case(
            (
                sa_start < 0,
                f(sa_arg, sa_start),
            ),
            else_=f(sa_arg, sa_start + 1),
        )

    sa_length = t.translate(op.length)

    return sa.case(
        (
            sa_start < 0,
            f(sa_arg, sa_start, sa_length),
        ),
        else_=f(sa_arg, sa_start + 1, sa_length),
    )


def _capitalize(t: ExprTranslator, op: ops.Capitalize) -> ir.Expr:
    sa_arg = t.translate(op.arg)
    return sa.func.concat(
        sa.func.ucase(sa.func.left(sa_arg, 1)),
        sa.func.lcase(sa.func.substring(sa_arg, 2)),
    )


_truncate_formats = {
    's': '%Y-%m-%d %H:%i:%s',
    'm': '%Y-%m-%d %H:%i:00',
    'h': '%Y-%m-%d %H:00:00',
    'D': '%Y-%m-%d',
    # 'W': 'week',
    'M': '%Y-%m-01',
    'Y': '%Y-01-01',
}


def _truncate(
    t: ExprTranslator,
    op: Union[ops.TimestampTruncate, ops.DateTruncate, ops.TimeTruncate],
) -> ir.Expr:
    sa_arg = t.translate(op.arg)
    try:
        fmt = _truncate_formats[op.unit.short]
    except KeyError:
        raise com.UnsupportedOperationError(f'Unsupported truncate unit {op.unit}')
    return sa.func.date_format(sa_arg, fmt)


def _cast(t: ExprTranslator, op: ops.Cast) -> ir.Expr:
    arg = op.arg
    to = op.to
    sa_arg = t.translate(arg)
    sa_type = t.get_sqla_type(op.to)

    # specialize going from an integer type to a timestamp
    if isinstance(arg.output_dtype, dt.Integer) and isinstance(sa_type, sa.DateTime):
        return sa.func.convert_tz(sa.func.from_unixtime(sa_arg), 'SYSTEM', 'UTC')

    if arg.output_dtype.equals(dt.binary) and to.equals(dt.string):
        return sa.func.hex(sa_arg)

    if to.equals(dt.binary):
        #  decode yields a column of memoryview which is annoying to deal with
        # in pandas. CAST(expr AS BYTEA) is correct and returns byte strings.
        return sa.cast(sa_arg, sa.LargeBinary())

    return sa.cast(sa_arg, sa_type)


def _round(t: ExprTranslator, op: ops.Round) -> ir.Expr:
    sa_arg = t.translate(op.arg)
    if op.digits is None:
        sa_digits = 0
    else:
        sa_digits = t.translate(op.digits)
    return sa.func.round(sa_arg, sa_digits)


def _quantile(t: ExprTranslator, op: ops.Quantile) -> ir.Expr:
    if op.interpolation is not None:
        warnings.warn(
            f"`{t.__module__.rsplit(',', 1)[0]}` backend does not support the "
            '`interpolation` argument',
        )
    arg = op.arg
    if (where := op.where) is not None:
        arg = ops.Where(where, arg, None)
        raise com.OperationNotDefinedError(
            'SingleStoreDB does not support `where` argument',
        )
    return sa.func.approx_percentile(t.translate(arg), t.translate(op.quantile))


def _multi_quantile(t: ExprTranslator, op: ops.Quantile) -> ir.Expr:
    raise com.OperationNotDefinedError(
        'SingleStoreDB does not support multiple quantiles simultaneously',
    )


def _interval_from_integer(t: ExprTranslator, op: ops.IntervalFromInteger) -> ir.Expr:
    if op.unit.short in {'ms', 'ns'}:
        raise com.UnsupportedOperationError(
            f'SingloStoreDB does not allow operation with INTERVAL offset {op.unit}',
        )

    sa_arg = t.translate(op.arg)
    text_unit = op.output_dtype.resolution.upper()

    # XXX: Is there a better way to handle this? I.e. can we somehow use
    # the existing bind parameter produced by translate and reuse its name in
    # the string passed to sa.text?
    if isinstance(sa_arg, sa.sql.elements.BindParameter):
        return sa.text(f'INTERVAL :arg {text_unit}').bindparams(arg=sa_arg.value)
    return sa.text(f'INTERVAL {sa_arg} {text_unit}')


def _literal(t: ExprTranslator, op: ops.Literal) -> ir.Expr:
    if op.output_dtype.is_interval():
        if op.output_dtype.unit.short in {'ms', 'ns'}:
            raise com.UnsupportedOperationError(
                'SingleStoreDB does not allow operation '
                f'with INTERVAL offset {op.output_dtype.unit}',
            )
        text_unit = op.output_dtype.resolution.upper()
        sa_text = sa.text(f'INTERVAL :value {text_unit}')
        return sa_text.bindparams(value=op.value)

    elif op.output_dtype.is_binary():
        # the cast to BINARY is necessary here, otherwise the data come back as
        # Python strings
        #
        # This lets the database handle encoding rather than ibis
        return sa.cast(sa.literal(op.value), type_=sa.BINARY())

    value = op.value
    with contextlib.suppress(AttributeError):
        value = value.to_pydatetime()

    return sa.literal(value)


def _group_concat(t: ExprTranslator, op: ops.GroupConcat) -> ir.Expr:
    if op.where is not None:
        arg = t.translate(ops.Where(op.where, op.arg, ibis.NA))
    else:
        arg = t.translate(op.arg)
    sep = t.translate(op.sep)
    return sa.func.group_concat(arg.op('SEPARATOR')(sep))


def _string_find(t: ExprTranslator, op: ops.StringFind) -> ir.Expr:
    if op.end is not None:
        raise NotImplementedError('`end` not yet implemented')

    if op.start is not None:
        return (
            sa.func.locate(
                t.translate(op.substr),
                t.translate(op.arg),
                t.translate(op.start),
            )
            - 1
        )

    return sa.func.locate(t.translate(op.substr), t.translate(op.arg)) - 1


def _string_contains(t: ExprTranslator, op: ops.StringContains) -> ir.Expr:
    return (sa.func.locate(t.translate(op.needle), t.translate(op.haystack)) - 1) >= 0


def _approx_median(t: ExprTranslator, op: ops.ApproxMedian) -> ir.Expr:
    return sa.func.median(t.translate(op.arg))


def _regex_search(t: ExprTranslator, op: ops.RegexSearch) -> ir.Expr:
    args = (op.arg, op.pattern)
    return sa.type_coerce(
        sa.func.regexp_instr(*[t.translate(x) for x in args], sa.literal('g')) > 0,
        sa.BOOLEAN,
    )


def _regex_replace(t: ExprTranslator, op: ops.RegexReplace) -> ir.Expr:
    # TODO: Requires regexp_format='advanced'
    args = (op.arg, op.pattern, op.replacement)
    return sa.func.regexp_replace(*[t.translate(x) for x in args], sa.literal('g'))


def _regex_extract(t: ExprTranslator, op: ops.RegexExtract) -> ir.Expr:
    args = (op.arg, op.pattern, op.index)
    return sa.func.regexp_extract(*[t.translate(x) for x in args], sa.literal('g'))


def _json_get_item(t: ExprTranslator, op: ops.JSONGetItem) -> ir.Expr:
    return sa.func.json_extract_json(t.translate(op.arg), t.translate(op.index))


def _json_delete_key(
    t: ExprTranslator, op: functions.JSONDeleteKey,
) -> ir.Expr:
    return sa.func.json_delete_key(t.translate(op.arg), *map(t.translate, op.key_path))


def _json_extract_double(t: ExprTranslator, op: functions.JSONExtractDouble) -> ir.Expr:
    return sa.func.json_extract_double(
        t.translate(op.arg), *map(t.translate, op.key_path),
    )


def _json_extract_string(t: ExprTranslator, op: functions.JSONExtractString) -> ir.Expr:
    return sa.func.json_extract_string(
        t.translate(op.arg), *map(t.translate, op.key_path),
    )


def _json_extract_json(t: ExprTranslator, op: functions.JSONExtractJSON) -> ir.Expr:
    out = sa.func.json_extract_json(t.translate(op.arg), *map(t.translate, op.key_path))
    return out


def _json_extract_bigint(t: ExprTranslator, op: functions.JSONExtractBigint) -> ir.Expr:
    return sa.func.json_extract_bigint(
        t.translate(op.arg), *map(t.translate, op.key_path),
    )


def _json_keys(
    t: ExprTranslator, op: functions.JSONKeys,
) -> ir.Expr:
    return sa.func.json_keys(t.translate(op.arg), *map(t.translate, op.key_path))


def _json_set_double(t: ExprTranslator, op: functions.JSONSetDouble) -> ir.Expr:
    return sa.func.json_set_double(
        t.translate(op.arg), *map(t.translate, op.key_path), t.translate(op.value),
    )


def _json_set_string(t: ExprTranslator, op: functions.JSONSetString) -> ir.Expr:
    return sa.func.json_set_string(
        t.translate(op.arg), *map(t.translate, op.key_path), t.translate(op.value),
    )


def _json_set_json(t: ExprTranslator, op: functions.JSONSetJSON) -> ir.Expr:
    return sa.func.json_set_json(
        t.translate(op.arg), *map(t.translate, op.key_path), t.translate(op.value),
    )


def _json_set(t: ExprTranslator, op: functions.JSONSet) -> ir.Expr:
    return sa.func.json_set(
        t.translate(op.arg), *map(t.translate, op.key_path), t.translate(op.value),
    )


def _json_splice_double(t: ExprTranslator, op: functions.JSONSpliceDouble) -> ir.Expr:
    return sa.func.json_splice_double(
        t.translate(op.arg),
        t.translate(op.start),
        t.translate(op.length),
        *map(t.translate, op.values),
    )


def _json_splice_string(t: ExprTranslator, op: functions.JSONSpliceString) -> ir.Expr:
    return sa.func.json_splice_string(
        t.translate(op.arg),
        t.translate(op.start),
        t.translate(op.length),
        *map(t.translate, op.values),
    )


def _json_splice_json(t: ExprTranslator, op: functions.JSONSpliceJSON) -> ir.Expr:
    return sa.func.json_splice_json(
        t.translate(op.arg),
        t.translate(op.start),
        t.translate(op.length),
        *map(t.translate, op.values),
    )


def _json_splice(t: ExprTranslator, op: functions.JSONSplice) -> ir.Expr:
    return sa.func.json_splice_json(
        t.translate(op.arg),
        t.translate(op.start),
        t.translate(op.length),
        *map(t.translate, op.values),
    )


def _to_number(t: ExprTranslator, op: functions.ToNumber) -> ir.Expr:
    if op.format_string is None:
        return sa.func.to_number(t.translate(op.arg))
    return sa.func.to_number(t.translate(op.arg), t.translate(op.format_string))


def _vector_sort(
    t: ExprTranslator,
    op: functions.VectorSort,
    dtype: Optional[str] = '',
) -> ir.Expr:
    func = sa.func.vector_sort
    if dtype:
        func = getattr(sa.func, 'vector_sort_' + dtype)
    if op.direction is None:
        return func(t.translate(op.arg))
    return func(t.translate(op.arg), t.translate(op.direction))


def _trunc(t: ExprTranslator, op: functions.Trunc) -> ir.Expr:
    if op.decimals is None:
        return sa.func.trunc(t.translate(op.arg))
    return sa.func.trunc(t.translate(op.arg), t.translate(op.decimals))


def _from_unixtime(t: ExprTranslator, expr: ir.Expr) -> ir.Expr:
    return sa.func.from_unixtime(t.translate(expr), sa.literal('YYYY-MM-DD HH:MI:SS'))


def _timestamp_from_unix(t: ExprTranslator, op: ops.TimestampFromUNIX) -> ir.Expr:
    val, unit = op.args
    val = util.convert_unit(val, unit.short, 's').to_expr().cast('int32').op()
    arg = _from_unixtime(t, val)
    return sa.cast(arg, sa.TIMESTAMP)


def _timestamp_from_ymdhms(t: ExprTranslator, op: ops.TimestampFromYMDHMS) -> ir.Expr:
    return sa.func.to_timestamp(
        sa.func.concat(
            sa.func.lpad(t.translate(op.year), 4, '0'),
            sa.literal('-'),
            sa.func.lpad(t.translate(op.month), 2, '0'),
            sa.literal('-'),
            sa.func.lpad(t.translate(op.day), 2, '0'),
            sa.literal(' '),
            sa.func.lpad(t.translate(op.hours), 2, '0'),
            sa.literal(':'),
            sa.func.lpad(t.translate(op.minutes), 2, '0'),
            sa.literal(':'),
            sa.func.lpad(t.translate(op.seconds), 2, '0'),
        ),
        'YYYY-MM-DD HH24:MI:SS',
    )


def _date_from_ymd(t: ExprTranslator, op: ops.DateFromYMD) -> ir.Expr:
    return sa.func.date(
        sa.func.concat(
            t.translate(op.year),
            sa.literal('-'),
            sa.func.lpad(t.translate(op.month), 2, '0'),
            sa.literal('-'),
            sa.func.lpad(t.translate(op.day), 2, '0'),
        ),
    )


def _time_from_hms(t: ExprTranslator, op: ops.TimeFromHMS) -> ir.Expr:
    return sa.func.time(
        sa.func.concat(
            sa.func.lpad(t.translate(op.hours), 2, '0'),
            sa.literal(':'),
            sa.func.lpad(t.translate(op.minutes), 2, '0'),
            sa.literal(':'),
            sa.func.lpad(t.translate(op.seconds), 2, '0'),
        ),
    )


operation_registry.update(
    {
        ops.Literal: _literal,
        ops.Cast: _cast,
        ops.TryCast: _cast,
        ops.IfNull: fixed_arity(sa.func.ifnull, 2),
        # static checks are not happy with using "if" as a property
        ops.Where: fixed_arity(getattr(sa.func, 'if'), 3),
        # strings
        ops.Substring: _substr,
        ops.StringFind: _string_find,
        ops.StringContains: _string_contains,
        ops.Capitalize: _capitalize,
        # ops.FindInSet: (
        #     lambda t, op: (
        #         sa.func.find_in_set(
        #             t.translate(op.needle),
        #             sa.func.concat_ws(',', *map(t.translate, op.values)),
        #         )
        #         - 1
        #     )
        # ),
        # LIKE in singlestoredb is case insensitive
        ops.StartsWith: fixed_arity(
            lambda arg, start: sa.type_coerce(
                arg.op('LIKE BINARY')(sa.func.concat(start, '%')),
                sa.BOOLEAN(),
            ),
            2,
        ),
        ops.EndsWith: fixed_arity(
            lambda arg, end: sa.type_coerce(
                arg.op('LIKE BINARY')(sa.func.concat('%', end)),
                sa.BOOLEAN(),
            ),
            2,
        ),
        ops.RegexSearch: _regex_search,
        ops.RegexReplace: _regex_replace,
        # ops.RegexExtract: _regex_extract,
        # math
        ops.Log: fixed_arity(lambda arg, base: sa.func.log(base, arg), 2),
        ops.Log2: unary(sa.func.log2),
        ops.Log10: unary(sa.func.log10),
        ops.Round: _round,
        ops.Quantile: _quantile,
        ops.MultiQuantile: _multi_quantile,
        functions.BitCount: unary(sa.func.bit_count),
        functions.Conv: fixed_arity(sa.func.conv, 3),
        functions.Sigmoid: unary(sa.func.sigmoid),
        functions.ToNumber: _to_number,
        functions.Trunc: _trunc,
        functions.Truncate: fixed_arity(sa.func.truncate, 2),
        # dates and times
        ops.DateAdd: fixed_arity(operator.add, 2),
        ops.DateSub: fixed_arity(operator.sub, 2),
        ops.DateDiff: fixed_arity(sa.func.datediff, 2),
        ops.TimestampAdd: fixed_arity(operator.add, 2),
        ops.TimestampSub: fixed_arity(operator.sub, 2),
        ops.TimestampDiff: fixed_arity(
            lambda left, right: sa.func.timestampdiff(sa.text('SECOND'), right, left),
            2,
        ),
        ops.StringToTimestamp: fixed_arity(
            lambda arg, format_str: sa.cast(
                sa.func.str_to_date(arg, format_str), sa.TIMESTAMP,
            ),
            2,
        ),
        ops.DateTruncate: _truncate,
        ops.TimestampTruncate: _truncate,
        ops.IntervalFromInteger: _interval_from_integer,
        ops.Strftime: fixed_arity(sa.func.date_format, 2),
        ops.ExtractDayOfYear: unary(sa.func.dayofyear),
        ops.ExtractEpochSeconds: unary(sa.func.UNIX_TIMESTAMP),
        ops.ExtractWeekOfYear: unary(sa.func.weekofyear),
        ops.ExtractMillisecond: fixed_arity(
            lambda arg: sa.func.floor(sa.extract('microsecond', arg) / 1000),
            1,
        ),
        ops.TimestampNow: fixed_arity(sa.func.now, 0),
        ops.DateFromYMD: _date_from_ymd,
        ops.TimeFromHMS: _time_from_hms,
        ops.TimestampFromYMDHMS: _timestamp_from_ymdhms,
        # ops.TimestampFromUNIX: _timestamp_from_unix,
        # reductions
        ops.ApproxMedian: _approx_median,
        # others
        ops.GroupConcat: _group_concat,
        ops.DayOfWeekIndex: fixed_arity(
            lambda arg: (sa.func.dayofweek(arg) + 5) % 7,
            1,
        ),
        ops.DayOfWeekName: fixed_arity(lambda arg: sa.func.dayname(arg), 1),
        ops.JSONGetItem: _json_get_item,
        ops.ToJSONArray: lambda t, op: sa.cast(
            sa.case(
                (
                    sa.func.json_get_type(t.translate(op.arg)) == 'array',
                    sa.func.to_json(t.translate(op.arg)),
                ),
                else_=sa.null(),
            ),
            sa.JSON,
        ),
        ops.ToJSONMap: lambda t, op: sa.cast(
            sa.case(
                (
                    sa.func.json_get_type(t.translate(op.arg)) == 'object',
                    sa.func.to_json(t.translate(op.arg)),
                ),
                else_=sa.null(),
            ),
            sa.JSON,
        ),
        ops.Strip: lambda t, op: sa.func.ltrim(sa.func.rtrim(t.translate(op.arg))),
        ops.LStrip: unary(sa.func.ltrim),
        ops.RStrip: unary(sa.func.rtrim),
        # vector functions
        functions.DotProduct: fixed_arity(sa.func.dot_product, 2),
        functions.DotProductI8: fixed_arity(sa.func.dot_product_i8, 2),
        functions.DotProductI16: fixed_arity(sa.func.dot_product_i16, 2),
        functions.DotProductI32: fixed_arity(sa.func.dot_product_i32, 2),
        functions.DotProductI64: fixed_arity(sa.func.dot_product_i64, 2),
        functions.DotProductF32: fixed_arity(sa.func.dot_product_f32, 2),
        functions.DotProductF64: fixed_arity(sa.func.dot_product_f64, 2),
        functions.EuclideanDistance: fixed_arity(sa.func.euclidean_distance, 2),
        functions.EuclideanDistanceI8: fixed_arity(sa.func.euclidean_distance_i8, 2),
        functions.EuclideanDistanceI16: fixed_arity(sa.func.euclidean_distance_i16, 2),
        functions.EuclideanDistanceI32: fixed_arity(sa.func.euclidean_distance_i32, 2),
        functions.EuclideanDistanceI64: fixed_arity(sa.func.euclidean_distance_i64, 2),
        functions.EuclideanDistanceF32: fixed_arity(sa.func.euclidean_distance_f32, 2),
        functions.EuclideanDistanceF64: fixed_arity(sa.func.euclidean_distance_f64, 2),
        functions.JSONArrayPack: unary(sa.func.json_array_pack),
        functions.JSONArrayPackI8: unary(sa.func.json_array_pack_i8),
        functions.JSONArrayPackI16: unary(sa.func.json_array_pack_i16),
        functions.JSONArrayPackI32: unary(sa.func.json_array_pack_i32),
        functions.JSONArrayPackI64: unary(sa.func.json_array_pack_i64),
        functions.JSONArrayPackF32: unary(sa.func.json_array_pack_f32),
        functions.JSONArrayPackF64: unary(sa.func.json_array_pack_f64),
        functions.JSONArrayUnpack: unary(sa.func.json_array_unpack),
        functions.JSONArrayUnpackI8: unary(sa.func.json_array_unpack_i8),
        functions.JSONArrayUnpackI16: unary(sa.func.json_array_unpack_i16),
        functions.JSONArrayUnpackI32: unary(sa.func.json_array_unpack_i32),
        functions.JSONArrayUnpackI64: unary(sa.func.json_array_unpack_i64),
        functions.JSONArrayUnpackF32: unary(sa.func.json_array_unpack_f32),
        functions.JSONArrayUnpackF64: unary(sa.func.json_array_unpack_f64),
        functions.ScalarVectorMul: fixed_arity(
            lambda arg, n: sa.func.scalar_vector_mul(n, arg), 2,
        ),
        functions.ScalarVectorMulI8: fixed_arity(
            lambda arg, n: sa.func.scalar_vector_mul_i8(n, arg), 2,
        ),
        functions.ScalarVectorMulI16: fixed_arity(
            lambda arg, n: sa.func.scalar_vector_mul_i16(n, arg), 2,
        ),
        functions.ScalarVectorMulI32: fixed_arity(
            lambda arg, n: sa.func.scalar_vector_mul_i32(n, arg), 2,
        ),
        functions.ScalarVectorMulI64: fixed_arity(
            lambda arg, n: sa.func.scalar_vector_mul_i64(n, arg), 2,
        ),
        functions.ScalarVectorMulF32: fixed_arity(
            lambda arg, n: sa.func.scalar_vector_mul_f32(n, arg), 2,
        ),
        functions.ScalarVectorMulF64: fixed_arity(
            lambda arg, n: sa.func.scalar_vector_mul_f64(n, arg), 2,
        ),
        functions.VectorAdd: fixed_arity(sa.func.vector_add, 2),
        functions.VectorAddI8: fixed_arity(sa.func.vector_add_i8, 2),
        functions.VectorAddI16: fixed_arity(sa.func.vector_add_i16, 2),
        functions.VectorAddI32: fixed_arity(sa.func.vector_add_i32, 2),
        functions.VectorAddI64: fixed_arity(sa.func.vector_add_i64, 2),
        functions.VectorAddF32: fixed_arity(sa.func.vector_add_f32, 2),
        functions.VectorAddF64: fixed_arity(sa.func.vector_add_f64, 2),
        functions.VectorElementsSum: unary(sa.func.vector_elements_sum),
        functions.VectorElementsSumI8: unary(sa.func.vector_elements_sum_i8),
        functions.VectorElementsSumI16: unary(sa.func.vector_elements_sum_i16),
        functions.VectorElementsSumI32: unary(sa.func.vector_elements_sum_i32),
        functions.VectorElementsSumI64: unary(sa.func.vector_elements_sum_i64),
        functions.VectorElementsSumF32: unary(sa.func.vector_elements_sum_f32),
        functions.VectorElementsSumF64: unary(sa.func.vector_elements_sum_f64),
        functions.VectorKthElement: fixed_arity(sa.func.vector_kth_element, 2),
        functions.VectorKthElementI8: fixed_arity(sa.func.vector_kth_element_i8, 2),
        functions.VectorKthElementI16: fixed_arity(sa.func.vector_kth_element_i16, 2),
        functions.VectorKthElementI32: fixed_arity(sa.func.vector_kth_element_i32, 2),
        functions.VectorKthElementI64: fixed_arity(sa.func.vector_kth_element_i64, 2),
        functions.VectorKthElementF32: fixed_arity(sa.func.vector_kth_element_f32, 2),
        functions.VectorKthElementF64: fixed_arity(sa.func.vector_kth_element_f64, 2),
        functions.VectorMul: fixed_arity(sa.func.vector_mul, 2),
        functions.VectorMulI8: fixed_arity(sa.func.vector_mul_i8, 2),
        functions.VectorMulI16: fixed_arity(sa.func.vector_mul_i16, 2),
        functions.VectorMulI32: fixed_arity(sa.func.vector_mul_i32, 2),
        functions.VectorMulI64: fixed_arity(sa.func.vector_mul_i64, 2),
        functions.VectorMulF32: fixed_arity(sa.func.vector_mul_f32, 2),
        functions.VectorMulF64: fixed_arity(sa.func.vector_mul_f64, 2),
        functions.VectorNumElements: unary(sa.func.vector_num_elements),
        functions.VectorNumElementsI8: unary(sa.func.vector_num_elements_i8),
        functions.VectorNumElementsI16: unary(sa.func.vector_num_elements_i16),
        functions.VectorNumElementsI32: unary(sa.func.vector_num_elements_i32),
        functions.VectorNumElementsI64: unary(sa.func.vector_num_elements_i64),
        functions.VectorNumElementsF32: unary(sa.func.vector_num_elements_f32),
        functions.VectorNumElementsF64: unary(sa.func.vector_num_elements_f64),
        functions.VectorSort: functools.partial(_vector_sort),
        functions.VectorSortI8: functools.partial(_vector_sort, dtype='i8'),
        functions.VectorSortI16: functools.partial(_vector_sort, dtype='i16'),
        functions.VectorSortI32: functools.partial(_vector_sort, dtype='i32'),
        functions.VectorSortI64: functools.partial(_vector_sort, dtype='i64'),
        functions.VectorSortF32: functools.partial(_vector_sort, dtype='f32'),
        functions.VectorSortF64: functools.partial(_vector_sort, dtype='f64'),
        functions.VectorSub: fixed_arity(sa.func.vector_sub, 2),
        functions.VectorSubI8: fixed_arity(sa.func.vector_sub_i8, 2),
        functions.VectorSubI16: fixed_arity(sa.func.vector_sub_i16, 2),
        functions.VectorSubI32: fixed_arity(sa.func.vector_sub_i32, 2),
        functions.VectorSubI64: fixed_arity(sa.func.vector_sub_i64, 2),
        functions.VectorSubF32: fixed_arity(sa.func.vector_sub_f32, 2),
        functions.VectorSubF64: fixed_arity(sa.func.vector_sub_f64, 2),
        functions.VectorSum: reduction(sa.func.vector_sum),
        functions.VectorSumI8: reduction(sa.func.vector_sum_i8),
        functions.VectorSumI16: reduction(sa.func.vector_sum_i16),
        functions.VectorSumI32: reduction(sa.func.vector_sum_i32),
        functions.VectorSumI64: reduction(sa.func.vector_sum_i64),
        functions.VectorSumF32: reduction(sa.func.vector_sum_f32),
        functions.VectorSumF64: reduction(sa.func.vector_sum_f64),
        functions.VectorSubvector: fixed_arity(sa.func.vector_subvector, 3),
        functions.VectorSubvectorI8: fixed_arity(sa.func.vector_subvector_i8, 3),
        functions.VectorSubvectorI16: fixed_arity(sa.func.vector_subvector_i16, 3),
        functions.VectorSubvectorI32: fixed_arity(sa.func.vector_subvector_i32, 3),
        functions.VectorSubvectorI64: fixed_arity(sa.func.vector_subvector_i64, 3),
        functions.VectorSubvectorF32: fixed_arity(sa.func.vector_subvector_f32, 3),
        functions.VectorSubvectorF64: fixed_arity(sa.func.vector_subvector_f64, 3),
        # json functions
        functions.JSONArrayContainsDouble: fixed_arity(
            sa.func.json_array_contains_double, 2,
        ),
        functions.JSONArrayContainsString: fixed_arity(
            sa.func.json_array_contains_string, 2,
        ),
        functions.JSONArrayContainsJSON: fixed_arity(
            sa.func.json_array_contains_json, 2,
        ),
        functions.JSONArrayContains: fixed_arity(
            sa.func.json_array_contains_json, 2,
        ),
        functions.JSONArrayPushDouble: fixed_arity(sa.func.json_array_push_double, 2),
        functions.JSONArrayPushString: fixed_arity(sa.func.json_array_push_string, 2),
        functions.JSONArrayPushJSON: fixed_arity(sa.func.json_array_push_json, 2),
        functions.JSONArrayPush: fixed_arity(sa.func.json_array_push_json, 2),
        functions.JSONDeleteKey: _json_delete_key,
        functions.JSONExtractDouble: _json_extract_double,
        functions.JSONExtractString: _json_extract_string,
        functions.JSONExtractJSON: _json_extract_json,
        functions.JSONExtractBigint: _json_extract_bigint,
        functions.JSONGetType: unary(sa.func.json_get_type),
        functions.JSONExcludeMask: fixed_arity(sa.func.json_exclude_mask, 2),
        functions.JSONIncludeMask: fixed_arity(sa.func.json_include_mask, 2),
        functions.JSONKeys: _json_keys,
        functions.JSONLength: unary(sa.func.json_length),
        functions.JSONPretty: unary(sa.func.json_pretty),
        functions.JSONSetDouble: _json_set_double,
        functions.JSONSetString: _json_set_string,
        functions.JSONSetJSON: _json_set_json,
        functions.JSONSet: _json_set,
        functions.JSONSpliceDouble: _json_splice_double,
        functions.JSONSpliceString: _json_splice_string,
        functions.JSONSpliceJSON: _json_splice_json,
        functions.JSONSplice: _json_splice,
        functions.Hex: unary(sa.func.hex),
        functions.Unhex: unary(sa.func.unhex),
    },
)

# Operations defined by base class, but not supported by SingleStoreDB
_invalid_operations = {
    ops.CumulativeAll,
    ops.CumulativeAny,
    ops.CumulativeMax,
    ops.CumulativeMean,
    ops.CumulativeMin,
    ops.CumulativeSum,
    ops.NTile,
    ops.Repeat,
}

operation_registry = {
    k: v for k, v in operation_registry.items() if k not in _invalid_operations
}


def _describe_table(
    self: Any,
    percentiles: Sequence[float] = [0.25, 0.5, 0.75],
    include: Optional[Sequence[Any]] = None,
    exclude: Optional[Sequence[Any]] = None,
    datetime_is_numeric: bool = False,
    stats: Optional[Sequence[str]] = None,
    by: Optional[Sequence[Union[str, ir.Expr]]] = None,
    having: Optional[Sequence[ir.Expr]] = None,
) -> ir.TableExpr:
    """
    Compute descriptive statistics.

    Note that since all statistics are stored in columns, there can
    be mixed data types represented in some columns such as ``min``,
    ``max``, ``top``, etc. Because of this, any column that contains
    the actual value of a column has a string type and the value
    must be parsed from the return value.

    Parameters
    ----------
    percentiles : list-like of floats, optional
        The percentiles to compute. Values should be between 0 and 1.
    include : 'all' or list-like of dtypes, optional
        By default, only numeric columns are included in the results.
        Specify 'all' to include all variables. A list of data types
        can also be specified. For example, ``numpy.number`` will
        select all numerics, ``numpy.object`` will select all
        objects types (i.e., strings and binary).
    exclude : list-like of dtypes, optional
        A list of data types to exclude.
    datetime_is_numeric : bool, optional
        Whether to treat datetimes as numeric. This affects the statistics
        computed for the column.
    stats : list-like of strings, optional
        Statistics to include in the output. By default, statistics are
        chosen based on data type. For numerics, ``count``, ``mean``,
        ``median``, ``std``, ``var``, ``min``, ``pct``, and ``max``
        are computed. For strings and binary, ``count``, ``unique``,
        ``min``, ``max``, ``top``, and ``freq`` are computed. For datetimes,
        ``count``, ``unique``, ``min``, ``max``, ``top``, and ``freq``
        are computed. If ``datetime_is_numeric`` is specified, all numeric
        computations are done for datetimes as well.
    by : list-like of strings or expressions, optional
        Group by variables
    having : list-like of expressions, optional
        Expressions that filter based on aggregate value

    Returns
    -------
    TableExpr

    """
    stat_map = dict(unique='nunique', median='approx_median')
    type_map = dict(
        std='double',
        mean='double',
        var='double',
        count='int64',
        unique='int64',
    )

    # Valid stats for each type
    num_stats = set('count mean median std var min pct max'.split())
    str_stats = set('count unique min max top freq'.split())
    dt_stats = set('count unique min max top freq'.split())

    if datetime_is_numeric:
        dt_stats.union(num_stats)

    # Compute includes and excludes

    if include is None:
        include = ['number']
    elif not isinstance(include, Sequence) or isinstance(include, str):
        include = [include]

    if exclude is None:
        exclude = []
    elif not isinstance(exclude, Sequence) or isinstance(exclude, str):
        exclude = [exclude]

    include_map = {
        np.datetime64: 'datetime',
        'datetime64': 'datetime',
        np.timedelta64: 'timedelta',
        'timedelta64': 'timedelta',
        'category': 'object',
        'datetimetz': 'datetime',
        'datetime64[ns, tz]': 'datetime',
        int: 'int',
        'i': 'int',
        float: 'float',
        'f': 'float',
        str: 'string',
        'str': 'string',
        bytes: 'binary',
        'bytes': 'binary',
        'O': 'object',
    }

    include_set = set([include_map.get(x, x) for x in include])
    exclude_set = set([include_map.get(x, x) for x in exclude])
    include_set = include_set.difference(exclude_set)

    include_vars = []
    for name, dtype in self.schema().items():
        if 'all' in include_set or 'one' in include_set:
            include_vars.append((name, dtype))
        elif 'object' in include_set and isinstance(dtype, (dt.String, dt.Binary)):
            include_vars.append((name, dtype))
        elif 'string' in include_set and isinstance(dtype, dt.String):
            include_vars.append((name, dtype))
        elif 'number' in include_set and isinstance(
            dtype, (dt.Integer, dt.Floating, dt.Decimal),
        ):
            include_vars.append((name, dtype))
        elif (
            'number' in include_set
            and datetime_is_numeric
            and isinstance(dtype, (dt.Timestamp, dt.Date, dt.Time, dt.Interval))
        ):
            include_vars.append((name, dtype))
        elif 'float' in include_set and isinstance(dtype, (dt.Floating, dt.Decimal)):
            include_vars.append((name, dtype))
        elif 'int' in include_set and isinstance(dtype, dt.Integer):
            include_vars.append((name, dtype))
        elif 'datetime' in include_set and isinstance(
            dtype, (dt.Timestamp, dt.Date, dt.Time),
        ):
            include_vars.append((name, dtype))
        elif 'timedelta' in include_set and isinstance(dtype, dt.Interval):
            include_vars.append((name, dtype))
        elif 'bytes' in include_set and isinstance(dtype, dt.Binary):
            include_vars.append((name, dtype))

    if not include_vars:
        raise ValueError(
            'No variables selected. Use `include=` or `exclude=` '
            'to select valid data types.',
        )

    # Compute stats list
    type_stats: Set[str] = set()
    for name, dtype in include_vars:
        if isinstance(dtype, (dt.Timestamp, dt.Date, dt.Time, dt.Interval)):
            type_stats = type_stats.union(dt_stats)
        elif isinstance(dtype, (dt.String, dt.Binary, dt.Array, dt.Enum)):
            type_stats = type_stats.union(str_stats)
        else:
            type_stats = type_stats.union(num_stats)

    stats = [
        x
        for x in [
            'count',
            'unique',
            'mean',
            'median',
            'std',
            'var',
            'min',
            'pct',
            'max',
            'top',
            'freq',
        ]
        if x in type_stats and (stats is None or x in stats)
    ]

    union = []

    # Build table expression
    if 'one' in include_set:
        out_type = str(include_vars[0][-1])
    else:
        out_type = 'string'
    for name, dtype in include_vars:
        if 'one' in include_set:
            agg = {}
        else:
            agg = {'name': ibis.literal(name)}
        freq = None
        for stat in stats:
            mthd_name = stat_map.get(stat, stat)
            # Expand 'pct' to percentile values
            if stat == 'pct':
                for pct in percentiles:
                    pct_label = str(int(pct * 100)) + '%'
                    if hasattr(self[name], 'quantile'):
                        agg[pct_label] = self[name].quantile(pct).cast(out_type)
                    else:
                        agg[pct_label] = ibis.null().cast(out_type)
            elif stat in ['top', 'freq']:
                freq = (
                    self[name]
                    .topk(1)
                    .to_aggregation(metric_name='freq')[
                        lambda x: x[name].cast(out_type).name('top'),
                        'freq',
                    ]
                )
            else:
                mthd = getattr(self[name], mthd_name, None)
                if mthd is None:
                    agg[stat] = ibis.null().cast(type_map.get(stat, out_type))
                elif stat == 'median' and isinstance(dtype, dt.String):
                    agg[stat] = ibis.null().cast(type_map.get(stat, out_type))
                else:
                    agg[stat] = mthd().cast(type_map.get(stat, out_type))

        union.append(self.aggregate(**agg, by=by, having=having))

        if freq is not None:
            union[-1] = union[-1].cross_join(freq)

    return ibis.union(*union)


ir.Table.describe = _describe_table


def _grouped_describe(self: Any, *args: Any, **kwargs: Any) -> pd.DataFrame:
    """
    Compute descriptive statistics.

    Note that since all statistics are stored in columns, there can
    be mixed data types represented in some columns such as ``min``,
    ``max``, ``top``, etc. Because of this, any column that contains
    the actual value of a column has a string type and the value
    must be parsed from the return value.

    Parameters
    ----------
    percentiles : list-like of floats, optional
        The percentiles to compute. Values should be between 0 and 1.
    include : 'all' or list-like of dtypes, optional
        By default, only numeric columns are included in the results.
        Specify 'all' to include all variables. A list of data types
        can also be specified. For example, ``numpy.number`` will
        select all numerics, ``numpy.object`` will select all
        objects types (i.e., strings and binary).
    exclude : list-like of dtypes, optional
        A list of data types to exclude.
    datetime_is_numeric : bool, optional
        Whether to treat datetimes as numeric. This affects the statistics
        computed for the column.
    stats : list-like of strings, optional
        Statistics to include in the output. By default, statistics are
        chosen based on data type. For numerics, ``count``, ``mean``,
        ``median``, ``std``, ``var``, ``min``, ``pct``, and ``max``
        are computed. For strings and binary, ``count``, ``unique``,
        ``min``, ``max``, ``top``, and ``freq`` are computed. For datetimes,
        ``count``, ``unique``, ``min``, ``max``, ``top``, and ``freq``
        are computed. If ``datetime_is_numeric`` is specified, all numeric
        computations are done for datetimes as well.

    Returns
    -------
    TableExpr

    """
    return (
        self.table.describe(*args, by=self.by, having=self._having, **kwargs)
        .select(lambda x: x)
        .sort_by([x.get_name() for x in self.by])
    )


gby.GroupedTable.describe = _grouped_describe


def _describe_column(
    self: Any,
    percentiles: Sequence[float] = [0.25, 0.5, 0.75],
    datetime_is_numeric: bool = False,
    stats: Optional[Sequence[str]] = None,
) -> ir.TableExpr:
    """
    Compute descriptive statistics.

    Parameters
    ----------
    percentiles : list-like of floats, optional
        The percentiles to compute. Values should be between 0 and 1.
    datetime_is_numeric : bool, optional
        Whether to treat datetimes as numeric. This affects the statistics
        computed for the column.
    stats : list-like of strings, optional
        Statistics to include in the output. By default, statistics are
        chosen based on data type. For numerics, ``count``, ``mean``,
        ``median``, ``std``, ``var``, ``min``, ``pct``, and ``max``
        are computed. For strings and binary, ``count``, ``unique``,
        ``min``, ``max``, ``top``, and ``freq`` are computed. For datetimes,
        ``count``, ``unique``, ``min``, ``max``, ``top``, and ``freq``
        are computed. If ``datetime_is_numeric`` is specified, all numeric
        computations are done for datetimes as well.

    Returns
    -------
    TableExpr

    """
    return _describe_table(
        self.to_projection(),
        percentiles=percentiles,
        include='one',
        exclude=None,
        datetime_is_numeric=datetime_is_numeric,
        stats=stats,
    )


ir.AnyColumn.describe = _describe_column


def _head_column(self: ir.AnyColumn, n: int = 5) -> ir.Expr:
    """Return first ``n`` row values."""
    return self.to_projection().head(n)[0]


ir.AnyColumn.head = _head_column


def _drop_duplicates(
    self: Any,
    subset: Optional[Union[str, Sequence[str]]] = None,
    keep: str = 'first',
    order_by: Optional[Union[str, Sequence[str], ir.Expr]] = None,
) -> ir.Table:
    """
    Drop rows with duplicate values.

    Parameters
    ----------
    subset : str or list-of-strs, optional
        The name or names of columns to compare for duplicate values
    keep : str, optional
        Which duplicate to keep: first or last
    order_by : str or Expr, optional
        The sort order for the duplicates to determine first or last

    Returns
    -------
    Table

    """
    if keep not in ['first', 'last']:
        raise ValueError('`keep` must be "first" or "last"')

    if order_by is None:
        raise ValueError(
            '`order_by` must be specified for ' '`keep="first"` or `keep="last"`',
        )

    if subset is None:
        subset = list(self.columns)
    elif isinstance(subset, str):
        subset = [subset]

    # TODO: Workaround for bug in Ibis where ibis.desc/ibis.asc can't
    #       be used in an order_by.
    def order_func(obj: Any) -> Any:
        func = ibis.asc if keep == 'first' else ibis.desc
        return func(order_by).resolve(obj)

    return (
        self.group_by(subset)
        .order_by(order_func)
        .mutate(**{'^ROW^ORDER^': ibis.row_number()})
        .filter(lambda x: x['^ROW^ORDER^'] == 0)
        .drop('^ROW^ORDER^')
    )


ir.Table.drop_duplicates = _drop_duplicates



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Tuple

import ibis.expr.datatypes as dt
import sqlalchemy.types as sat
import sqlalchemy_singlestoredb as singlestoredb
from ibis.backends.base.sql.alchemy.datatypes import AlchemyType
from ibis.backends.base.sql.alchemy.datatypes import UUID
from ibis.common.exceptions import UnsupportedOperationError

# binary character set
# used to distinguish blob binary vs blob text
MY_CHARSET_BIN = 63


def _type_from_cursor_info(descr: Tuple[Any, ...], field: Any) -> dt.DataType:
    """Construct an ibis type from SingleStoreDB field descr and field result metadata.

    This method is complex because the SingleStoreDB protocol is complex.

    Types are not encoded in a self contained way, meaning you need
    multiple pieces of information coming from the result set metadata to
    determine the most precise type for a field. Even then, the decoding is
    not high fidelity in some cases: UUIDs for example are decoded as
    strings, because the protocol does not appear to preserve the logical
    type, only the physical type.
    """
    _, type_code, _, _, field_length, scale, _, _, _ = descr
    flags = _FieldFlags(field.flags)
    typename = _type_codes.get(type_code)
    if typename is None:
        raise NotImplementedError(
            f'SingleStoreDB type code {type_code:d} is not supported',
        )

    kwargs = {}

    if typename in ('DECIMAL', 'NEWDECIMAL'):
        precision = _decimal_length_to_precision(
            length=field_length,
            scale=scale,
            is_unsigned=flags.is_unsigned,
        )
        typ = partial(_type_mapping[typename], precision=precision, scale=scale)

    elif typename == 'BIT':
        typ = dt.int64

    elif typename == 'YEAR':
        typ = dt.uint8

    elif typename == 'TIMESTAMP':
        typ = dt.Timestamp
        kwargs['timezone'] = 'UTC'
        # if scale > 0:
        #     kwargs['scale'] = scale

    elif typename == 'DATETIME':
        typ = dt.Timestamp
        # if scale > 0:
        #     kwargs['scale'] = scale

    elif flags.is_set:
        # sets are limited to strings
        typ = dt.Array(dt.string)

    elif flags.is_unsigned and type_code in _num_types:
        typ = getattr(dt, f'U{_type_mapping[typename].__name__}')

    elif type_code in _char_types:
        # binary text
        if field.charsetnr == MY_CHARSET_BIN:
            typ = dt.Binary
        else:
            typ = dt.String

    else:
        typ = _type_mapping[typename]

    # projection columns are always nullable
    return typ(nullable=True, **kwargs)


# ported from my_decimal.h:my_decimal_length_to_precision in mariadb
def _decimal_length_to_precision(*, length: int, scale: int, is_unsigned: bool) -> int:
    return length - (scale > 0) - (not (is_unsigned or not length))


_num_types = {1, 2, 3, 4, 5, 8, 9}
_char_types = {15, 249, 250, 251, 252, 253, 254, 255}

_type_codes = {
    0: 'DECIMAL',
    1: 'TINY',
    2: 'SHORT',
    3: 'LONG',
    4: 'FLOAT',
    5: 'DOUBLE',
    6: 'NULL',
    7: 'TIMESTAMP',
    8: 'LONGLONG',
    9: 'INT24',
    10: 'DATE',
    11: 'TIME',
    12: 'DATETIME',
    13: 'YEAR',
    15: 'VARCHAR',
    16: 'BIT',
    245: 'JSON',
    246: 'NEWDECIMAL',
    247: 'ENUM',
    248: 'SET',
    249: 'TINY_BLOB',
    250: 'MEDIUM_BLOB',
    251: 'LONG_BLOB',
    252: 'BLOB',
    253: 'VAR_STRING',
    254: 'STRING',
    255: 'GEOMETRY',
}


_type_mapping = {
    'DECIMAL': dt.Decimal,
    'TINY': dt.Int8,
    'SHORT': dt.Int16,
    'LONG': dt.Int32,
    'FLOAT': dt.Float32,
    'DOUBLE': dt.Float64,
    'NULL': dt.Null,
    'TIMESTAMP': lambda nullable: dt.Timestamp(timezone='UTC', nullable=nullable),
    'LONGLONG': dt.Int64,
    'INT24': dt.Int32,
    'DATE': dt.Date,
    'TIME': dt.Time,
    'DATETIME': dt.Timestamp,
    'YEAR': dt.Int8,
    'VARCHAR': dt.String,
    'JSON': dt.JSON,
    'NEWDECIMAL': dt.Decimal,
    'ENUM': dt.String,
    'SET': lambda nullable: dt.Array(dt.string, nullable=nullable),
    'TINY_BLOB': dt.Binary,
    'MEDIUM_BLOB': dt.Binary,
    'LONG_BLOB': dt.Binary,
    'BLOB': dt.Binary,
    'VAR_STRING': dt.String,
    'STRING': dt.String,
    'GEOMETRY': dt.Geometry,
}


class _FieldFlags:
    """Flags used to disambiguate field types.

    Gaps in the flag numbers are because we do not map in flags that are
    of no use in determining the field's type, such as whether the field
    is a primary key or not.
    """

    UNSIGNED = 1 << 5
    SET = 1 << 11
    NUM = 1 << 15

    __slots__ = ('value',)

    def __init__(self, value: int) -> None:
        self.value = value

    @property
    def is_unsigned(self) -> bool:
        return (self.UNSIGNED & self.value) != 0

    @property
    def is_set(self) -> bool:
        return (self.SET & self.value) != 0

    @property
    def is_num(self) -> bool:
        return (self.NUM & self.value) != 0


class SingleStoreDBDateTime(singlestoredb.DATETIME):
    """Custom DATETIME type for SingleStoreDB that handles zero values."""

    def result_processor(self, *_: Any) -> Any:
        return lambda v: None if v == '0000-00-00 00:00:00' else v


_to_singlestoredb_types = {
    dt.Boolean: singlestoredb.BOOLEAN,
    dt.Int8: singlestoredb.TINYINT,
    dt.Int16: singlestoredb.SMALLINT,
    dt.Int32: singlestoredb.INTEGER,
    dt.Int64: singlestoredb.BIGINT,
    dt.Float16: singlestoredb.FLOAT,
    dt.Float32: singlestoredb.FLOAT,
    dt.Float64: singlestoredb.DOUBLE,
    dt.String: singlestoredb.TEXT,
    dt.JSON: singlestoredb.JSON,
    dt.Timestamp: SingleStoreDBDateTime,
}

_from_singlestoredb_types = {
    singlestoredb.BIGINT: dt.Int64,
    singlestoredb.BINARY: dt.Binary,
    singlestoredb.BLOB: dt.Binary,
    singlestoredb.BOOLEAN: dt.Boolean,
    singlestoredb.DATETIME: dt.Timestamp,
    singlestoredb.DOUBLE: dt.Float64,
    singlestoredb.FLOAT: dt.Float32,
    singlestoredb.INTEGER: dt.Int32,
    singlestoredb.JSON: dt.JSON,
    singlestoredb.LONGBLOB: dt.Binary,
    singlestoredb.LONGTEXT: dt.String,
    singlestoredb.MEDIUMBLOB: dt.Binary,
    singlestoredb.MEDIUMINT: dt.Int32,
    singlestoredb.MEDIUMTEXT: dt.String,
    singlestoredb.REAL: dt.Float64,
    singlestoredb.SMALLINT: dt.Int16,
    singlestoredb.TEXT: dt.String,
    singlestoredb.DATE: dt.Date,
    singlestoredb.TINYBLOB: dt.Binary,
    singlestoredb.TINYINT: dt.Int8,
    singlestoredb.TINYTEXT: dt.String,
    singlestoredb.VARBINARY: dt.Binary,
    singlestoredb.VARCHAR: dt.String,
    singlestoredb.ENUM: dt.String,
    singlestoredb.CHAR: dt.String,
    singlestoredb.TIME: dt.Time,
    singlestoredb.YEAR: dt.UInt8,
    SingleStoreDBDateTime: dt.Timestamp,
    UUID: dt.String,
}

_unsigned_int_map = {
    dt.Int8: dt.UInt8,
    dt.Int16: dt.UInt16,
    dt.Int32: dt.UInt32,
    dt.Int64: dt.UInt64,
}


class SingleStoreDBType(AlchemyType):

    dialect = 'singlestoredb'

    @classmethod
    def from_ibis(cls, dtype: Any) -> Any:
        try:
            out = _to_singlestoredb_types[type(dtype)]
            if isinstance(dtype, dt.Timestamp):
                if getattr(dtype, 'timezone', 'UTC') not in [False, None, 'UTC', 'utc']:
                    raise UnsupportedOperationError('All timestamps must be in UTC')
                scale = getattr(dtype, 'scale', 0)
                if scale:
                    out = out(fsp=scale)
                else:
                    out = out()
            elif isinstance(dtype, dt.Decimal):
                out = out(
                    precision=getattr(dtype, 'precision', None) or None,
                    scale=getattr(dtype, 'scale', None) or None,
                )
            elif isinstance(dtype, dt.UnsignedInteger):
                out = out(unsigned=True)
            return out
        except KeyError:
            return super().from_ibis(dtype)

    @classmethod
    def to_ibis(cls, typ: Any, nullable: bool = True) -> Any:
        if isinstance(typ, (sat.NUMERIC, singlestoredb.NUMERIC, singlestoredb.DECIMAL)):
            return dt.Decimal(typ.precision or 10, typ.scale or 0, nullable=nullable)
        elif isinstance(typ, singlestoredb.BIT):
            return dt.Int64(nullable=nullable)
        elif isinstance(
            typ, (singlestoredb.TIMESTAMP, singlestoredb.DATETIME, SingleStoreDBDateTime),
        ):
            if getattr(typ, 'timezone', 'UTC') not in [False, None, 'UTC', 'utc']:
                raise UnsupportedOperationError('All timestamps must be in UTC')
            kwargs = dict(
                # scale=getattr(typ, 'scale', None),
                timezone=None if isinstance(typ, singlestoredb.DATETIME) else 'UTC',
            )
            return dt.Timestamp(nullable=nullable, **kwargs)
        elif isinstance(typ, singlestoredb.SET):
            return dt.Set(dt.string, nullable=nullable)
        elif dtype := _from_singlestoredb_types[type(typ)]:
            if getattr(typ, 'unsigned', False):
                dtype = _unsigned_int_map.get(dtype, dtype)
            return dtype(nullable=nullable)
        else:
            return super().to_ibis(typ, nullable=nullable)
